"""
wire has the basic extended types useful for creating logic.

Types defined in this file include:

- `WireVector` -- the base class for ordered collections of wires

- `Input` -- a wire vector that receives an input for a block

- `Output` -- a wire vector that defines an output for a block

- `Const` -- a wire vector fed by a constant

- `Register` -- a wire vector that is latched each cycle
"""

from __future__ import annotations

import numbers
import re
import traceback
from typing import Union

from pyrtl import core  # needed for _setting_keep_wirevector_call_stack
from pyrtl.core import Block, LogicNet, _NameIndexer, working_block
from pyrtl.pyrtlexceptions import PyrtlError, PyrtlInternalError

# ----------------------------------------------------------------
#        ___  __  ___  __   __
#  \  / |__  /  `  |  /  \ |__)
#   \/  |___ \__,  |  \__/ |  \
#
_wvIndexer = _NameIndexer("tmp")
_constIndexer = _NameIndexer("const_")


def _reset_wire_indexers():
    global _wvIndexer, _constIndexer
    _wvIndexer = _NameIndexer("tmp")
    _constIndexer = _NameIndexer("const_")


def next_tempvar_name(name=""):
    if name == "":  # sadly regex checks are sometimes too slow
        wire_name = _wvIndexer.make_valid_string()
        callpoint = core._get_useful_callpoint_name()
        if callpoint:  # returns none if debug mode is false
            filename, lineno = callpoint
            safename = re.sub(
                r"[\W]+", "", filename
            )  # strip out non alphanumeric characters
            wire_name += f"_{safename}_line{lineno}"
        return wire_name
    if name.lower() in ["clk", "clock"]:
        msg = "Clock signals should never be explicit"
        raise PyrtlError(msg)
    return name


class WireVector:
    """The main class for describing the connections between operators.

    ``WireVectors`` act much like a list of wires, except that there is no "contained"
    type, each slice of a ``WireVector`` is itself a ``WireVector`` (even if it just
    contains a single "bit" of information). The least significant bit of the wire is at
    index ``0`` and normal list slicing syntax applies (i.e. ``myvector[0:5]`` makes a
    new vector from the bottom ``5`` bits of ``myvector``, ``myvector[-1]`` takes the
    most significant bit, and ``myvector[-4:]`` takes the ``4`` most significant bits).

    ==============  ===========  ===========================================================  ===================
    Operation       Syntax       Function                                                     Documentation
    ==============  ===========  ===========================================================  ===================
    Addition        ``a + b``    Creates an *unsigned* adder, returns ``WireVector``          :meth:`__add__`
    Subtraction     ``a - b``    Creates an *unsigned* subtracter, returns ``WireVector``     :meth:`__sub__`
    Multiplication  ``a * b``    Creates an *unsigned* multiplier, returns ``WireVector``     :meth:`__mul__`
    Xor             ``a ^ b``    Bitwise XOR, returns ``WireVector``                          :meth:`__xor__`
    Or              ``a | b``    Bitwise OR, returns ``WireVector``                           :meth:`__or__`
    And             ``a & b``    Bitwise AND, returns ``WireVector``                          :meth:`__and__`
    Invert          ``~a``       Bitwise invert, returns ``WireVector``                       :meth:`__invert__`
    Less Than       ``a < b``    Unsigned less than, return 1-bit ``WireVector``              :meth:`__lt__`
    Less or Eq.     ``a <= b``   Unsigned less than or equal to, return 1-bit ``WireVector``  :meth:`__le__`
    Greater Than    ``a > b``    Unsigned greater than, return 1-bit ``WireVector``           :meth:`__gt__`
    Greater or Eq.  ``a >= b``   Unsigned greater or equal to, return 1-bit ``WireVector``    :meth:`__ge__`
    Equality        ``a == b``   Hardware to check equality, return 1-bit ``WireVector``      :meth:`__eq__`
    Not Equal       ``a != b``   Inverted equality check, return 1-bit ``WireVector``         :meth:`__ne__`
    Bitwidth        ``len(a)``   Return bitwidth of the ``WireVector``                        :meth:`__len__`
    Assignment      ``a <<= b``  Connect from b to a (see Note below)                         :meth:`__ilshift__`
    Bit Slice       ``a[3:6]``   Selects bits from ``WireVector``, in this case bits 3,4,5    :meth:`__getitem__`
    ==============  ===========  ===========================================================  ===================

    .. note::
        ``<<=`` is how you "drive" an already created wire with an existing wire. If you
        were to do ``a = b`` it would lose the old value of ``a`` and simply overwrite
        it with a new value, in this case with a reference to ``WireVector`` ``b``. In
        contrast ``a <<= b`` does not overwrite ``a``, but simply wires the two
        together.

    .. _wirevector_coercion:

    ``WireVector`` Coercion
    -----------------------

    Most PyRTL functions that accept ``WireVectors`` as arguments will also accept any
    type that :func:`as_wires` can coerce to ``WireVector``. Examples include
    :class:`int`, :class:`bool`, and :class:`str`.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    :class:`int` will be coerced to an *unsigned* :class:`Const` ``WireVector`` with
    the minimum bitwidth required for the integer. In the following example, a
    2-bit :class:`Const` is implicitly created for ``2``::

        >>> input = pyrtl.Input(name="input", bitwidth=8)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= input + 2

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": 3})
        >>> sim.inspect("output")
        5

    .. doctest only::
        >>> pyrtl.reset_working_block()

    :class:`bool` will be coerced to a :class:`Const` with :attr:`bitwidth` ``1``. In
    the following example, a 1-bit :class:`Const` is implicitly created for ``True``::

        >>> input = pyrtl.Input(name="input", bitwidth=1)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= input ^ True

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": False})
        >>> sim.inspect("output")
        1

    .. doctest only::
        >>> pyrtl.reset_working_block()

    :class:`str` will be interpreted as a `Verilog-style string constant
    <https://en.wikipedia.org/wiki/Verilog#Definition_of_constants>`_. In the following
    example, a 4-bit :class:`Const` is implicitly created for ``"4'hf"``::

        >>> input = pyrtl.Input(name="input", bitwidth=8)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= input & "4'hf"

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": 0xab})
        >>> hex(sim.inspect("output"))
        '0xb'

    .. _wirevector_equality:

    ``WireVector`` Equality
    -----------------------

    :meth:`WireVector.__eq__` generates logic that dynamically reports if two wires
    carry the same values. :meth:`WireVector.__eq__` returns a 1-bit ``WireVector``, not
    a ``bool``, and attempting to convert a ``WireVector`` to a ``bool`` raises a
    :class:`PyrtlError`. This behavior is incompatible with `Python's data model
    <https://docs.python.org/3/reference/expressions.html#value-comparisons>`_, which
    can cause problems.

    For example, you *can not* statically check if two ``WireVectors`` are equal with
    ``==``. Statically checking for ``WireVector`` equality can be useful while
    constructing or analyzing circuits::

        >>> w1 = pyrtl.WireVector(name="w1", bitwidth=1)
        >>> w2 = pyrtl.WireVector(name="w2", bitwidth=2)
        >>> if w1 == w2:
        ...     print('same')
        ...
        Traceback (most recent call last):
        ...
        pyrtl.pyrtlexceptions.PyrtlError: cannot convert WireVector to compile-time boolean...

    The error about converting ``WireVector`` to ``bool`` results from Python attempting
    to convert the 1-bit ``WireVector`` returned by :meth:`__eq__` to ``True`` or
    ``False`` while evaluating the ``if`` statement's condition.

    Instead, you *can* statically check if two ``WireVectors`` refer to the same object
    with ``is``::

        >>> w1 is not w2
        True
        >>> temp = w1
        >>> temp is w1
        True
        >>> temp is w2
        False

    Be careful when using Python features that depend on ``==`` with ``WireVectors``.
    This often comes up when checking if a ``WireVector`` is in a :class:`list` with
    ``in``, which does not work because ``in`` falls back on checking each item in the
    :class:`list` for equality with ``==``::

        >>> l = [w1]
        >>> w2 in l
        Traceback (most recent call last):
        ...
        pyrtl.pyrtlexceptions.PyrtlError: cannot convert WireVector to compile-time
        boolean...

    Most other :class:`list` operations work, so you can store ``WireVectors`` in a
    :class:`list` if you avoid using the ``in`` operator::

        >>> len(l)
        1
        >>> l[0] is w1
        True
        >>> [(w.name, w.bitwidth) for w in l]
        [('w1', 1)]

    ``WireVectors`` define a standard ``__hash__`` method, so if you need to check if a
    ``WireVector`` is in a container, use a :class:`set` or :class:`dict`. This works
    because these containers use ``__hash__`` to skip unnecessary equality checks::

        >>> s = {w1}
        >>> w1 in s
        True
        >>> w2 in s
        False

        >>> d = {w1: 'hello'}
        >>> w1 in d
        True
        >>> w2 in d
        False
        >>> d[w1]
        'hello'

    """  # noqa: E501

    bitwidth: int
    """The wire's bitwidth.

    ``WireVectors`` can be constructed without specifying a ``bitwidth``. These
    ``WireVectors`` will have a ``bitwidth`` of ``None`` until they infer a ``bitwidth``
    from an ``<<=`` assignment.

    ``bitwidth`` is equivalent to :meth:`__len__`, except that :meth:`__len__` raises an
    exception when ``bitwidth`` is ``None``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> w = pyrtl.WireVector()
        >>> w.bitwidth
        None
        >>> w <<= pyrtl.Const(val=42, bitwidth=6)
        >>> len(w)
        6
    """

    block: Block
    """The ``Block`` that this ``WireVector`` belongs to."""

    # "code" is a static variable used when output as string.
    # Each class inheriting from WireVector should overload accordingly
    _code = "W"

    def __init__(
        self, bitwidth: int | None = None, name: str = "", block: Block = None
    ):
        """Construct a generic ``WireVector``.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Examples::

            >>> # Visible in Simulation traces as "data".
            >>> data = pyrtl.WireVector(bitwidth=8, name="data")
            >>> data.name
            'data'

            >>> # `ctrl` is assigned a temporary name, and will not be visible in
            >>> # Simulation traces by default.
            >>> ctrl = pyrtl.WireVector(bitwidth=1)
            >>> ctrl.name.startswith("tmp")
            True

            >>> # `temp` is a temporary with bitwidth specified later.
            >>> temp = pyrtl.WireVector()
            >>> temp.bitwidth is None
            True
            >>> # `temp` infers a bitwidth of 8 from `data`.
            >>> temp <<= data
            >>> temp.bitwidth
            8

        :param bitwidth: If no ``bitwidth`` is provided, it will be set to the minimum
            number of bits needed to represent this wire.
        :param block: The :class:`Block` under which the wire should be placed.
            Defaults to the :ref:`working_block`.
        :param name: The name of the wire. Must be unique. If empty, a name will be
            autogenerated. If non-empty, the wire's value can be inspected with
            :meth:`Simulation.inspect`, and this wire will appear in traces generated
            by :meth:`SimulationTrace.render_trace`.
        """
        self._name = None

        # used only to verify the one to one relationship of wires and blocks
        self._block = working_block(block)
        self.name = next_tempvar_name(name)
        self._validate_bitwidth(bitwidth)

        if core._setting_keep_wirevector_call_stack:
            self.init_call_stack = traceback.format_stack()

    @property
    def name(self) -> str:
        """A property holding the name of the ``WireVector``.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        The name can be read or written. Examples::

            >>> a = WireVector(name="foo", bitwidth=1)
            >>> a.name
            'foo'
            >>> a.name = "mywire"
            >>> a.name
            'mywire'
        """
        return self._name

    @name.setter
    def name(self, value: str):
        if not isinstance(value, str):
            msg = "WireVector names must be strings"
            raise PyrtlError(msg)
        self._block.wirevector_by_name.pop(self._name, None)
        self._name = value
        self._block.add_wirevector(self)

    def __hash__(self):
        return id(self)

    def __str__(self):
        """A string representation of the wire in 'name/bitwidth code' form."""
        return f"{self.name}/{self.bitwidth}{self._code}"

    def __repr__(self):
        return str(self)

    def _validate_bitwidth(self, bitwidth):
        if bitwidth is not None:
            if not isinstance(bitwidth, numbers.Integral):
                msg = (
                    "bitwidth must be from type int or unspecified, instead "
                    f'"{bitwidth}" was passed of type {type(bitwidth)}'
                )
                raise PyrtlError(msg)
            if bitwidth == 0:
                msg = "bitwidth must be greater than or equal to 1"
                raise PyrtlError(msg)
            if bitwidth < 0:
                msg = "you are trying a negative bitwidth? awesome but wrong"
                raise PyrtlError(msg)
        self.bitwidth = bitwidth

    def _build(self, other: WireVectorLike):
        # Actually create and add WireVector to logic block This might be called
        # immediately from ilshift, or delayed from conditional assignment
        net = LogicNet(op="w", op_param=None, args=(other,), dests=(self,))
        working_block().add_net(net)

    def _prepare_for_assignment(self, rhs):
        # Convert right-hand-side to wires and propagate bitwidth if necessary
        from pyrtl.corecircuits import as_wires

        rhs = as_wires(rhs, bitwidth=self.bitwidth)
        if self.bitwidth is None:
            self.bitwidth = rhs.bitwidth
        return rhs

    def __ilshift__(self, other: WireVectorLike):
        """Wire assignment operator (assign ``other`` to ``self``).

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> input = pyrtl.Input(bitwidth=8, name="input")
            >>> output = pyrtl.WireVector(name="output")

            >>> output <<= input

            >>> sim = pyrtl.Simulation()
            >>> sim.step(provided_inputs={"input": 42})
            >>> sim.inspect("output")
            42
        """
        other = self._prepare_for_assignment(other)
        self._build(other)
        return self

    def __ior__(self, other: WireVectorLike):
        """Conditional assignment operator (only valid under Conditional Update)."""
        from pyrtl.conditional import _build, currently_under_condition

        if not self.bitwidth:
            msg = (
                "Conditional assignment only defined on WireVectors with pre-defined "
                "bitwidths"
            )
            raise PyrtlError(msg)
        other = self._prepare_for_assignment(other)
        if currently_under_condition():
            _build(self, other)
        else:
            self._build(other)
        return self

    def _two_var_op(self, other, op):
        from pyrtl.corecircuits import as_wires, match_bitwidth

        # convert constants if necessary
        a, b = self, as_wires(other)
        a, b = match_bitwidth(a, b)
        resultlen = len(a)  # both are the same length now

        # some operations actually create more or less bits
        if op in "+-":
            resultlen += 1  # extra bit required for carry
        elif op == "*":
            resultlen = resultlen * 2  # more bits needed for mult
        elif op in "<>=":
            resultlen = 1

        s = WireVector(bitwidth=resultlen)
        net = LogicNet(op=op, op_param=None, args=(a, b), dests=(s,))
        working_block().add_net(net)
        return s

    def __bool__(self):
        """Use of a WireVector in a statement like "a or b" is forbidden."""
        # python provides no way to overload these logical operations, and thus they
        # are very much not likely to be doing the thing that the programmer would be
        # expecting.
        msg = (
            "cannot convert WireVector to compile-time boolean.  This error often "
            'happens when you attempt to use WireVectors with "==" or something that '
            'calls "__eq__", such as when you test if a WireVector is "in" something'
        )
        raise PyrtlError(msg)

    def __and__(self, other: WireVectorLike) -> WireVector:
        """Returns the result of bitwise ANDing ``self`` and ``other``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=0b11, bitwidth=2)
            >>> five = pyrtl.Const(val=0b101, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= three & five
            >>> output.bitwidth
            4

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> sim.inspect("output")
            1

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A ``WireVector`` containing the result of bitwise ANDing ``self`` and
                 ``other``. The returned ``WireVector`` has the same :attr:`bitwidth` as
                 the longer of the two input ``WireVectors``.
        """
        return self._two_var_op(other, "&")

    def __rand__(self, other: WireVectorLike):
        return self._two_var_op(other, "&")

    def __iand__(self, other: WireVectorLike):
        msg = "error, operation not allowed on WireVectors"
        raise PyrtlError(msg)

    def __or__(self, other: WireVectorLike) -> WireVector:
        """Returns the result of bitwise ORing ``self`` and ``other``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=0b11, bitwidth=2)
            >>> five = pyrtl.Const(val=0b101, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= three | five
            >>> output.bitwidth
            4

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> bin(sim.inspect("output"))
            '0b111'

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A ``WireVector`` containing the result of bitwise ORing ``self`` and
                 ``other``. The returned ``WireVector`` has the same :attr:`bitwidth` as
                 the longer of the two input ``WireVectors``.
        """
        return self._two_var_op(other, "|")

    def __ror__(self, other: WireVectorLike):
        return self._two_var_op(other, "|")

    # __ior__ used for conditional assignment above

    def __xor__(self, other: WireVectorLike) -> WireVector:
        """Returns the result of bitwise XORing ``self`` and ``other``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=0b11, bitwidth=2)
            >>> five = pyrtl.Const(val=0b101, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= three ^ five
            >>> output.bitwidth
            4

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> bin(sim.inspect("output"))
            '0b110'

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A ``WireVector`` containing the result of bitwise XORing ``self`` and
                 ``other``. The returned ``WireVector`` has the same :attr:`bitwidth` as
                 the longer of the two input ``WireVectors``.
        """
        return self._two_var_op(other, "^")

    def __rxor__(self, other: WireVectorLike):
        return self._two_var_op(other, "^")

    def __ixor__(self, other: WireVectorLike):
        msg = "error, operation not allowed on WireVectors"
        raise PyrtlError(msg)

    def __add__(self, other: WireVectorLike) -> WireVector:
        """Returns the result of adding ``self`` and ``other``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. WARNING::

            This addition operation is *unsigned*. Use :func:`signed_add` for signed
            addition.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=3, bitwidth=2)
            >>> five = pyrtl.Const(val=5, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= three + five
            >>> output.bitwidth
            5

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> sim.inspect("output")
            8

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A ``WireVector`` containing the result of adding ``self`` and
                 ``other``. The returned ``WireVector`` has a :attr:`bitwidth` equal to
                 the longer of the two input ``WireVectors``, plus one.
        """
        return self._two_var_op(other, "+")

    def __radd__(self, other: WireVectorLike):
        return self._two_var_op(other, "+")

    def __iadd__(self, other: WireVectorLike):
        msg = "error, operation not allowed on WireVectors"
        raise PyrtlError(msg)

    def __sub__(self, other: WireVectorLike) -> WireVector:
        """Returns the result of subtracting ``self`` and ``other``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. WARNING::

            This subtraction operation is *unsigned*. Use :func:`signed_sub` for signed
            subtraction.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=3, bitwidth=2)
            >>> five = pyrtl.Const(val=5, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= five - three
            >>> output.bitwidth
            5

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> sim.inspect("output")
            2

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A ``WireVector`` containing the result of subtracting ``self`` and
                 ``other``. The returned ``WireVector`` has a :attr:`bitwidth` equal to
                 the longer of the two input ``WireVectors``, plus one.
        """
        return self._two_var_op(other, "-")

    def __rsub__(self, other: WireVectorLike):
        from pyrtl.corecircuits import as_wires

        other = as_wires(other)  # '-' op is not symmetric
        return other._two_var_op(self, "-")

    def __isub__(self, other: WireVectorLike):
        msg = "error, operation not allowed on WireVectors"
        raise PyrtlError(msg)

    def __mul__(self, other: WireVectorLike) -> WireVector:
        """Returns the result of multiplying ``self`` and ``other``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. WARNING::

            This multiplication operation is *unsigned*. Use :func:`signed_mult` for
            signed multiplication.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=3, bitwidth=2)
            >>> five = pyrtl.Const(val=5, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= three * five
            >>> output.bitwidth
            8

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> sim.inspect("output")
            15

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A ``WireVector`` containing the result of multiplying ``self`` and
                 ``other``. The returned ``WireVector`` has a :attr:`bitwidth` equal to
                 twice the length of the longer input.
        """
        return self._two_var_op(other, "*")

    def __rmul__(self, other: WireVectorLike):
        return self._two_var_op(other, "*")

    def __imul__(self, other: WireVectorLike):
        msg = "error, operation not allowed on WireVectors"
        raise PyrtlError(msg)

    def __lt__(self, other: WireVectorLike) -> WireVector:
        """Checks if ``self`` is less than ``other``. Returns a one-bit ``WireVector``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. WARNING::

            This comparison is *unsigned*. Use :func:`signed_lt` for signed comparison.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=3, bitwidth=2)
            >>> five = pyrtl.Const(val=5, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= three < five
            >>> output.bitwidth
            1

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> sim.inspect("output")
            1

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A one-bit ``WireVector`` indicating if ``self`` is less than ``other``.
        """
        return self._two_var_op(other, "<")

    def __le__(self, other: WireVectorLike) -> WireVector:
        """Checks if ``self`` is less than or equal to ``other``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. WARNING::

            This comparison is *unsigned*. Use :func:`signed_le` for signed comparison.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=3, bitwidth=2)
            >>> five = pyrtl.Const(val=5, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= three <= five
            >>> output.bitwidth
            1

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> sim.inspect("output")
            1

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A one-bit ``WireVector`` indicating if ``self`` is less than or equal
                 to ``other``.
        """
        return ~self._two_var_op(other, ">")

    def __eq__(self, other: WireVectorLike) -> WireVector:
        """Checks if ``self`` is equal to ``other``. Returns a one-bit ``WireVector``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. WARNING::

            This definition of ``__eq__`` returns :class:`WireVector`, not
            :class:`bool`, which is not compatible with Python's data model, which can
            cause problems. See :ref:`wirevector_equality`.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=3, bitwidth=2)
            >>> five = pyrtl.Const(val=5, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= three == five
            >>> output.bitwidth
            1

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> sim.inspect("output")
            0

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A one-bit ``WireVector`` indicating if ``self`` is equal to ``other``.
        """
        return self._two_var_op(other, "=")

    def __ne__(self, other: WireVectorLike) -> WireVector:
        """Checks if ``self`` is not equal to ``other``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=3, bitwidth=2)
            >>> five = pyrtl.Const(val=5, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= three != five
            >>> output.bitwidth
            1

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> sim.inspect("output")
            1

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A one-bit ``WireVector`` indicating if ``self`` is not equal to
                 ``other``.
        """
        return ~self._two_var_op(other, "=")

    def __gt__(self, other: WireVectorLike) -> WireVector:
        """Checks if ``self`` is greater than ``other``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. WARNING::

            This comparison is *unsigned*. Use :func:`signed_gt` for signed comparison.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=3, bitwidth=2)
            >>> five = pyrtl.Const(val=5, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= three > five
            >>> output.bitwidth
            1

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> sim.inspect("output")
            0

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A one-bit ``WireVector`` indicating if ``self`` is greater than
                 ``other``.
        """
        return self._two_var_op(other, ">")

    def __ge__(self, other: WireVectorLike) -> WireVector:
        """Checks if ``self`` is greater than or equal to ``other``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. WARNING::

            This comparison is *unsigned*. Use :func:`signed_ge` for signed comparison.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=3, bitwidth=2)
            >>> five = pyrtl.Const(val=5, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= three >= five
            >>> output.bitwidth
            1

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> sim.inspect("output")
            0

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A one-bit ``WireVector`` indicating if ``self`` is greater than or
                 equal to ``other``.
        """
        return ~self._two_var_op(other, "<")

    def __invert__(self) -> WireVector:
        """Returns a ``WireVector`` containing the bitwise inversion of ``self``.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> five = pyrtl.Const(val=0b101, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= ~five
            >>> output.bitwidth
            4

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> bin(sim.inspect("output"))
            '0b1010'

        :return: A ``WireVector`` containing the result of bitwise inverting ``self``.
                 The returned ``WireVector`` has the same :attr:`bitwidth` as ``self``.
        """
        outwire = WireVector(bitwidth=len(self))
        net = LogicNet(op="~", op_param=None, args=(self,), dests=(outwire,))
        working_block().add_net(net)
        return outwire

    def __getitem__(self, item: int | slice) -> WireVector:
        """Returns a ``WireVector`` containing a subset of the wires in ``self``.

        There are two ways to retrieve ``WireVector`` subsets:

        1. By :class:`int` index, for example ``wire[2]``. This returns a ``WireVector``
           with :attr:`bitwidth` ``1``.

        2. By :class:`slice`, for example ``wire[3:5]``. Slicing uses the usual
           ``[start:stop:step]`` notation.

           1. If ``start`` is omitted, it defaults to index ``0``, the least significant
              bit.
           2. If ``stop`` is omitted, it defaults to ``bitwidth - 1``, the most
              significant bit.
           3. If ``step`` is omitted, it defaults to ``1``.

        ``bitwidth`` is added to negative ``start`` and ``stop`` indices, so negative
        indices count backwards from just beyond the most significant bit. Index
        ``bitwidth - 1`` and index ``-1`` both refer to the most significant bit.

        If ``step`` is negative, the wires will be returned in reverse order.

        Suppose we have a ``WireVector`` ``input``, with :attr:`bitwidth` 8::

            input = WireVector(name="input", bitwidth=8)

        We can access individual wires of ``input`` with integer indices::

            input[0]                   # Least significant bit.
            input[input.bitwidth - 1]  # Most significant bit
            input[-1]                  # Another name for the most significant bit.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> input = pyrtl.Input(name="input", bitwidth=8)
            >>> msb = pyrtl.Output(name="msb")

            >>> msb <<= input[-1]
            >>> msb.bitwidth
            1

            >>> sim = pyrtl.Simulation()
            >>> sim.step(provided_inputs={"input": 0b1000_0000})
            >>> sim.inspect("msb")
            1

        We can access contiguous subsets of ``input``'s wires with slices::

            input[2:6]  # Middle 4 bits.
            input[:4]   # Least significant 4 bits.
            input[-4:]  # Most significant 4 bits.

        .. doctest only::

            >>> pyrtl.reset_working_block()

        Example::

            >>> input = pyrtl.Input(name="input", bitwidth=8)
            >>> middle_bits = pyrtl.Output(name="middle_bits")

            >>> middle_bits <<= input[2:6]
            >>> middle_bits.bitwidth
            4

            >>> sim = pyrtl.Simulation()
            >>> sim.step(provided_inputs={"input": 0b0011_1100})
            >>> bin(sim.inspect("middle_bits"))
            '0b1111'

        We can skip some of ``input``'s wires with slices::

            input[::2]   # Even numbered wires: 0, 2, 4, 6.
            input[1::2]  # Odd numbered wires: 1, 3, 5, 7.

        .. doctest only::

            >>> pyrtl.reset_working_block()

        Example::

            >>> input = pyrtl.Input(name="input", bitwidth=8)
            >>> odd_bits = pyrtl.Output(name="odd_bits")

            >>> odd_bits <<= input[1::2]
            >>> odd_bits.bitwidth
            4

            >>> sim = pyrtl.Simulation()
            >>> sim.step(provided_inputs={"input": 0b1010_1010})
            >>> bin(sim.inspect("odd_bits"))
            '0b1111'

        We can reverse ``input``'s wires with slices::

            input[::-1]    # Reversed wires: 7, 6, 5, 4, 3, 2, 1, 0.
            input[-1::-2]  # Reversed odd wires: 7, 5, 3, 1.

        .. doctest only::

            >>> pyrtl.reset_working_block()

        Example::

            >>> input = pyrtl.Input(name="input", bitwidth=8)
            >>> reversed_bits = pyrtl.Output(name="reversed_bits")

            >>> reversed_bits <<= input[::-1]
            >>> reversed_bits.bitwidth
            8

            >>> sim = pyrtl.Simulation()
            >>> sim.step(provided_inputs={"input": 0b0000_1111})
            >>> bin(sim.inspect("reversed_bits"))
            '0b11110000'

            >>> sim.step(provided_inputs={"input": 0b1010_1010})
            >>> # `bin` omits the leading 0.
            >>> bin(sim.inspect("reversed_bits"))
            '0b1010101'

        :param item: If an :class:`int`, specifies the index of a single-bit wire to
            return. If a :class:`slice`, specifies a subset of wires to return, as
            ``start:stop:step``.
        :return: A ``WireVector`` containing the wires selected by ``item`` from
            ``self``.
        """
        if self.bitwidth is None:
            msg = "You cannot get a subset of a wire with no bitwidth"
            raise PyrtlError(msg)
        allindex = range(self.bitwidth)
        if isinstance(item, int):
            selectednums = (
                allindex[item],
            )  # this method handles negative numbers correctly
        else:  # slice
            selectednums = tuple(allindex[item])
        if not selectednums:
            msg = f"selection {item} must have at least one selected wire"
            raise PyrtlError(msg)
        outwire = WireVector(bitwidth=len(selectednums))
        net = LogicNet(op="s", op_param=selectednums, args=(self,), dests=(outwire,))
        working_block().add_net(net)
        return outwire

    def __lshift__(self, other: WireVectorLike):
        msg = (
            "Shifting using the << operator is not supported in PyRTL. If you are "
            "trying to select bits in a wire, use the indexing operator "
            "(wire[indexes]) instead.\n\nFor example: wire[2:9] selects the wires from "
            "index 2 to index 8 to make a new length 7 wire. \n\nIf you are really "
            'trying to *execution time* left shift you can use "shift_left_logical"'
        )
        raise PyrtlError(msg)

    def __rshift__(self, other: WireVectorLike):
        msg = (
            "Shifting using the >> operator is not supported in PyRTL. If you are "
            "trying to select bits in a wire, use the indexing operator "
            "(wire[indexes]) instead.\n\nFor example: wire[2:9] selects the wires from "
            "index 2 to index 8 to make a new length 7 wire. \n\nIf you are really "
            'trying to *execution time* right shift you can use "shift_right_logical" '
            'or "shift_right_arithmetic"'
        )
        raise PyrtlError(msg)

    def __mod__(self, other: WireVectorLike):
        msg = (
            "Masking with the % operator is not supported in PyRTL. Instead if you are "
            "trying to select bits in a wire, use the indexing operator "
            "(wire[indexes]) instead.\n\nFor example: wire[2:9] selects the wires from "
            "index 2 to index 8 to make a new length 7 wire."
        )
        raise PyrtlError(msg)

    def __len__(self) -> int:
        """Return the ``WireVector``'s :attr:`bitwidth`.

        ``WireVectors`` can be constructed without specifying a :attr:`bitwidth`. These
        ``WireVectors`` will have a :attr:`bitwidth` of ``None`` until they infer a
        :attr:`bitwidth` from an ``<<=`` assignment.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> w = pyrtl.WireVector()
            >>> len(w)
            Traceback (most recent call last):
            ...
            pyrtl.pyrtlexceptions.PyrtlError: length of WireVector not yet defined

            >>> w <<= pyrtl.Const(val=42, bitwidth=6)
            >>> len(w)
            6


        :return: Returns the length (:attr:`bitwidth`) of the ``WireVector``, in bits.

        :raises PyrtlError: If the :attr:`bitwidth` is not yet defined.
        """
        if self.bitwidth is None:
            msg = "length of WireVector not yet defined"
            raise PyrtlError(msg)
        return self.bitwidth

    def __enter__(self):
        """Use wires as contexts for conditional assignments."""
        from pyrtl.conditional import _push_condition

        _push_condition(self)

    def __exit__(self, *execinfo):
        from pyrtl.conditional import _pop_condition

        _pop_condition()

    # more functions for wires
    def nand(self, other: WireVectorLike) -> WireVector:
        """Returns the result of bitwise NANDing ``self`` and ``other``.

        If the inputs do not have the same :attr:`bitwidth`, the shorter input will be
        :meth:`zero_extended` to the longer input's :attr:`bitwidth`.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> three = pyrtl.Const(val=0b11, bitwidth=2)
            >>> five = pyrtl.Const(val=0b101, bitwidth=4)
            >>> output = pyrtl.Output(name="output")

            >>> output <<= three.nand(five)
            >>> output.bitwidth
            4

            >>> sim = pyrtl.Simulation()
            >>> sim.step()
            >>> bin(sim.inspect("output"))
            '0b1110'

        :param other: A ``WireVector``, or any type that can be coerced to
            ``WireVector`` by :func:`as_wires`.

        :return: A ``WireVector`` containing the result of bitwise NANDing ``self`` and
                 ``other``. The returned ``WireVector`` has the same :attr:`bitwidth` as
                 the longer of the two input ``WireVectors``.
        """
        return self._two_var_op(other, "n")

    @property
    def bitmask(self) -> int:
        """A property holding a bitmask of the same length as this ``WireVector``.

        ``bitmask`` is an :class:`int` with a number of bits set to 1 equal to the
        :attr:`bitwidth` of the ``WireVector``.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        It is often useful to "mask" an integer such that it fits in the number of bits
        of a ``WireVector``, so the ``bitmask`` property is provided as a convenience.
        Example::

            >>> w = pyrtl.WireVector(bitwidth=4)
            >>> bin(w.bitmask)
            '0b1111'
            >>> w.bitmask
            15
            >>> hex(0xabcd & w.bitmask)
            '0xd'
        """
        if "_bitmask" not in self.__dict__:
            self._bitmask = (1 << len(self)) - 1
        return self._bitmask

    def truncate(self, bitwidth: int) -> WireVector:
        """Return a copy of ``self`` with its most significant bits removed.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Truncation reduces :attr:`bitwidth` by removing the most significant bits from
        ``self``. Example::

            >>> input = pyrtl.Input(name="input", bitwidth=8)
            >>> output = pyrtl.Output(name="output", bitwidth=4)

            >>> output <<= input.truncate(bitwidth=output.bitwidth)

            >>> sim = pyrtl.Simulation()
            >>> sim.step(provided_inputs={"input": 0b0000_1111})
            >>> bin(sim.inspect("output"))
            '0b1111'

            >>> sim.step(provided_inputs={"input": 0b1100_1011})
            >>> bin(sim.inspect("output"))
            '0b1011'

        :param bitwidth: Number of bits to truncate to. This is the number of bits to
            keep.

        :return: A new ``WireVector`` equal to the original ``WireVector`` but truncated
                 to the specified :attr:`bitwidth`.

        :raises PyrtlError: If the :attr:`bitwidth` specified is larger than
            ``self.bitwidth``.
        """
        if not isinstance(bitwidth, int):
            msg = "Can only truncate to an integer number of bits"
            raise PyrtlError(msg)
        if bitwidth > self.bitwidth:
            msg = "Cannot truncate a WireVector to have more bits than it started with"
            raise PyrtlError(msg)
        return self[:bitwidth]

    def sign_extended(self, bitwidth) -> WireVector:
        """Return a sign-extended copy of ``self``.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Sign-extension increases :attr:`bitwidth` by adding copies of the most
        significant bit of ``self``. Example::

            >>> input = pyrtl.Input(name="input", bitwidth=1)
            >>> output = pyrtl.Output(name="output", bitwidth=4)

            >>> output <<= input.sign_extended(bitwidth=output.bitwidth)

            >>> sim = pyrtl.Simulation()
            >>> sim.step(provided_inputs={"input": 0})
            >>> bin(sim.inspect("output"))
            '0b0'

            >>> sim.step(provided_inputs={"input": 1})
            >>> bin(sim.inspect("output"))
            '0b1111'

        :return: A new ``WireVector`` equal to the original ``WireVector`` but
                 sign-extended to the specified :attr:`bitwidth`.

        :raises PyrtlError: If the :attr:`bitwidth` specified is smaller than
            ``self.bitwidth``.
        """
        return self._extend_with_bit(bitwidth, self[-1])

    def zero_extended(self, bitwidth) -> WireVector:
        """Return a zero-extended copy of ``self``.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Zero-extension increases :attr:`bitwidth` by adding zero-valued high bits to
        ``self``. Example::

            >>> input = pyrtl.Input(name="input", bitwidth=1)
            >>> output = pyrtl.Output(name="output", bitwidth=4)

            >>> output <<= input.zero_extended(bitwidth=output.bitwidth)

            >>> sim = pyrtl.Simulation()
            >>> sim.step(provided_inputs={"input": 0})
            >>> bin(sim.inspect("output"))
            '0b0'

            >>> sim.step(provided_inputs={"input": 1})
            >>> bin(sim.inspect("output"))
            '0b1'

        :return: A new ``WireVector`` equal to the original ``WireVector`` but
                 zero-extended to the specified :attr:`bitwidth`.

        :raises PyrtlError: If the :attr:`bitwidth` specified is smaller than
            ``self.bitwidth``.
        """
        return self._extend_with_bit(bitwidth, 0)

    def _extend_with_bit(self, bitwidth, extbit):
        numext = bitwidth - self.bitwidth
        if numext == 0:
            return self
        if numext < 0:
            msg = (
                "Neither zero_extended nor sign_extended can reduce the number of bits"
            )
            raise PyrtlError(msg)
        from pyrtl.corecircuits import concat

        if isinstance(extbit, int):
            extbit = Const(extbit, bitwidth=1)
        extvector = WireVector(bitwidth=numext)
        net = LogicNet(
            op="s", op_param=(0,) * numext, args=(extbit,), dests=(extvector,)
        )
        working_block().add_net(net)
        return concat(extvector, self)


WireVectorLike = Union[WireVector, int, str, bool]
"""Alias for types that can be coerced to ``WireVector`` by :func:`as_wires`."""


# -----------------------------------------------------------------------
#  ___     ___  ___       __   ___  __           ___  __  ___  __   __   __
# |__  \_/  |  |__  |\ | |  \ |__  |  \    \  / |__  /  `  |  /  \ |__) /__`
# |___ / \  |  |___ | \| |__/ |___ |__/     \/  |___ \__,  |  \__/ |  \ .__/
#
class Input(WireVector):
    """A ``WireVector`` placeholder for inputs to a :class:`Block`.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    ``Input`` ``WireVectors`` are placeholders for values provided during simulation.
    See :meth:`Simulation.step`'s ``provided_inputs`` argument. For example::

        >>> input = pyrtl.Input(name="input", bitwidth=8)
        >>> output = pyrtl.Output(name="output")
        >>> output <<= input + 2

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": 1})
        >>> sim.inspect("output")
        3

    .. doctest only::

        >>> pyrtl.reset_working_block()

    Attempting to assign an ``Input`` ``WireVector`` with the ``<<=`` or ``|=``
    operators will raise :class:`PyrtlError`::

        >>> input = pyrtl.Input(name="input", bitwidth=1)
        >>> input <<= True
        Traceback (most recent call last):
        ...
        pyrtl.pyrtlexceptions.PyrtlError: Connection ... attempted on Input
    """

    _code = "I"

    def __init__(
        self, bitwidth: int | None = None, name: str = "", block: Block = None
    ):
        super().__init__(bitwidth=bitwidth, name=name, block=block)

    def __ilshift__(self, _):
        """This is an illegal op for Inputs. They cannot be assigned to in this way"""
        msg = (
            "Connection using <<= operator attempted on Input. Inputs, such as "
            f'"{self.name}", cannot have values generated internally. aka they can\'t '
            "have other wires driving it"
        )
        raise PyrtlError(msg)

    def __ior__(self, _):
        """This is an illegal op for Inputs. They cannot be assigned to in this way"""
        msg = (
            "Connection using |= operator attempted on Input. Inputs, such as "
            f'"{self.name}", cannot have values generated internally. aka they can\'t '
            "have other wires driving it"
        )
        raise PyrtlError(msg)


class Output(WireVector):
    """A ``WireVector`` type denoting outputs of a :class:`Block`.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Attempting to use an ``Output`` ``WireVector`` as the input to any operation, such
    as ``__or__``, which implements bitwise or, will raise :class:`PyrtlInternalError`::

        >>> output = pyrtl.Output(name="output", bitwidth=1)
        >>> foo = output | 2
        Traceback (most recent call last):
        ...
        pyrtl.pyrtlexceptions.PyrtlInternalError: error, Outputs cannot be arguments for
        a net
    """

    _code = "O"

    def __init__(
        self, bitwidth: int | None = None, name: str = "", block: Block = None
    ):
        super().__init__(bitwidth, name, block)


class Const(WireVector):
    """A ``WireVector`` representation of a constant value.

    Converts from bool, integer, or Verilog-style strings to a constant of the specified
    :attr:`bitwidth`. If a positive integer is specified, the bitwidth can be inferred
    from the constant. If a negative integer is provided in the simulation, it is
    converted to a two's complement representation of the specified bitwidth.
    """

    _code = "C"

    def __init__(
        self,
        val: int | bool | str,
        bitwidth: int | None = None,
        name: str = "",
        signed: bool = False,
        block: Block = None,
    ):
        """Construct a constant implementation at initialization.

        .. WARNING::

            A constant generated with ``signed=True`` is still just a raw bitvector. All
            arthimetic on it is *unsigned* by default. The ``signed=True`` argument is
            only used for proper inference of WireVector size and certain bitwidth
            sanity checks, assuming a two's complement representation of the constants.
            For signed arithmetic, use the ``signed_*`` functions in the
            :ref:`extended_logic_and_arithmetic` section.

        :param val: The constant value. For details of how constants are coerced from
            int, bool, and strings (for Verilog constants), see documentation for
            :func:`infer_val_and_bitwidth`.
        :param bitwidth: The desired bitwidth of the resulting ``Const``.
        :param name: The name of the wire. Must be unique. If none is provided, one will
            be autogenerated
        :param signed: Specifies if bits should be used for two's complement.
        :param block: The block under which the wire should be placed. Defaults to the
            :ref:`working_block`.

        :raise PyrtlInternalError: If the :attr:`bitwidth` is too short to represent the
            specified constant.
        """
        self._validate_bitwidth(bitwidth)
        from pyrtl.helperfuncs import infer_val_and_bitwidth

        num, bitwidth = infer_val_and_bitwidth(val, bitwidth, signed)

        if num < 0:
            msg = "Const somehow evaluating to negative integer after checks"
            raise PyrtlInternalError(msg)
        if (num >> bitwidth) != 0:
            msg = (
                f"constant {num} returned by infer_val_and_bitwidth somehow not "
                f"fitting in {bitwidth} bits"
            )
            raise PyrtlInternalError(msg)

        name = name if name else _constIndexer.make_valid_string() + "_" + str(val)

        super().__init__(bitwidth=bitwidth, name=name, block=block)
        # add the member "val" to track the value of the constant
        self.val = num

    def __ilshift__(self, other: WireVectorLike):
        """
        This is an illegal op for Consts. Their value is set in the __init__ function
        """
        msg = (
            f'ConstWires, such as "{self.name}", should never be assigned to with "<<="'
        )
        raise PyrtlError(msg)

    def __ior__(self, _):
        """This is an illegal op for Consts. They cannot be assigned to in this way"""
        msg = (
            "Connection using |= operator attempted on Const. ConstWires, such as "
            f'"{self.name}", cannot have values generated internally. aka they cannot '
            "have other wires driving it"
        )
        raise PyrtlError(msg)


class Register(WireVector):
    """A WireVector with an embedded register state element.

    Registers only update their outputs on the rising edges of an implicit clock signal.
    The "value" in the current cycle can be accessed by referencing the Register itself.
    To set the value for the next cycle (after the next rising clock edge), set the
    :attr:`Register.next` property with the ``<<=`` operator.

    Registers reset to zero by default, and reside in the same clock domain.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> counter = pyrtl.Register(name="counter", bitwidth=2)
        >>> counter.next <<= counter + 1

        >>> sim = pyrtl.Simulation()
        >>> sim.step_multiple(nsteps=6)
        >>> sim.tracer.trace["counter"]
        [0, 1, 2, 3, 0, 1]

    This builds a zero-initialized 2-bit counter. The second line sets the counter's
    value in the next cycle (``counter.next``) to the counter's value in the current
    cycle (``counter``), plus one.
    """

    _code = "R"

    # When a register's next value is assigned, the following occurs:
    #
    # 1. The register's `.next` property is retrieved. Register.next returns an instance
    #    of Register._Next.
    # 2. __ilshift__ is invoked on the returned instance of Register._Next.
    #
    # So `reg.next <<= foo` effectively does the following:
    #
    #     reg.next = Register._Next(reg)
    #     reg.next.__ilshift__(reg, foo)
    #
    # The following behavior is expected:
    #
    #     reg.next <<= 5  # good
    #     a <<= reg       # good
    #     reg <<= 5       # error
    #     a <<= reg.next  # error
    #     reg.next = 5    # error
    class _Next:
        """Type returned by the ``Register.next`` property.

        This class allows unconditional assignments (``<<=``, ``__ilshift__``) and
        conditional assignments (``|=``, ``__ior__``) on ``Register.next``. Registers
        themselves do not support assignments, so ``Register.__ilshift__`` and
        ``Register.__ior__`` throw errors.

        ``__ilshift__`` and ``__ior__`` must both return ``self`` because::

            x <<= y

        is equivalent to::

            x = x.__ilshift__(y)

        Note how ``__ilshift__``'s return value is assigned to ``x``, see
        https://docs.python.org/3/library/operator.html#in-place-operators

        ``__ilshift__`` and ``__ior__`` both return ``self`` and Register's @next.setter
        checks that Register.next is assigned to an instance of _Next.
        """

        def __init__(self, reg):
            self.reg = reg

        def __ilshift__(self, other: WireVectorLike):
            from pyrtl.corecircuits import as_wires

            other = as_wires(other, bitwidth=self.reg.bitwidth)
            if self.reg.bitwidth is None:
                self.reg.bitwidth = other.bitwidth

            if self.reg.reg_in is not None:
                msg = "error, .next value should be set once and only once"
                raise PyrtlError(msg)
            self.reg._build(other)

            return self

        def __ior__(self, other: WireVectorLike):
            from pyrtl.conditional import _build
            from pyrtl.corecircuits import as_wires

            other = as_wires(other, bitwidth=self.reg.bitwidth)
            if not self.reg.bitwidth:
                msg = (
                    "Conditional assignment only defined on Registers with pre-defined "
                    "bitwidths"
                )
                raise PyrtlError(msg)

            if self.reg.reg_in is not None:
                msg = "error, .next value should be set once and only once"
                raise PyrtlError(msg)
            _build(self.reg, other)

            return self

        def __bool__(self):
            """Use of a _next in a statement like "a or b" is forbidden."""
            msg = (
                "cannot convert Register.next to compile-time boolean.  This error "
                'often happens when you attempt to use a Register.next with "==" or '
                'something that calls "__eq__", such as when you test if a '
                'Register.next is "in" something'
            )
            raise PyrtlError(msg)

    def __init__(
        self,
        bitwidth: int,
        name: str = "",
        reset_value: int | None = None,
        block: Block = None,
    ):
        """Construct a register.

        It is an error if the ``reset_value`` cannot fit into the specified bitwidth for
        this register.

        :param bitwidth: Number of bits to represent this register.
        :param name: The name of the register's current value (``reg``, not
            ``reg.next``). Must be unique. If none is provided, one will be
            autogenerated.
        :param reset_value: Value to initialize this register to during simulation and
            in any code (e.g. Verilog) that is exported. Defaults to 0. Can be
            overridden at simulation time.
        :param block: The block under which the wire should be placed. Defaults to the
            :ref:`working_block`.
        """
        from pyrtl.helperfuncs import infer_val_and_bitwidth

        super().__init__(bitwidth=bitwidth, name=name, block=block)
        self.reg_in = None  # wire vector setting self.next
        if reset_value is not None:
            reset_value, rst_bitwidth = infer_val_and_bitwidth(
                reset_value,
                bitwidth=bitwidth,
            )
            if rst_bitwidth > bitwidth:
                msg = (
                    f'reset_value "{reset_value}" cannot fit in the specified '
                    f"{bitwidth} bits for this register"
                )
                raise PyrtlError(msg)
        self.reset_value = reset_value

    @property
    def next(self):
        """Sets the Register's value for the next cycle (it is before the D-Latch)."""
        return Register._Next(self)

    def __ilshift__(self, other: WireVectorLike):
        msg = "error, you cannot set registers directly, net .next instead"
        raise PyrtlError(msg)

    def __ior__(self, other: WireVectorLike):
        msg = "error, you cannot set registers directly, net .next instead"
        raise PyrtlError(msg)

    @next.setter
    def next(self, other: WireVectorLike):
        if not isinstance(other, Register._Next):
            msg = 'error, .next should be set with "<<=" or "|=" operators'
            raise PyrtlError(msg)

    def _build(self, next):
        # Actually build the register. This happens immediately when setting the `next`
        # property. Under conditional assignment, register build is delayed until the
        # conditional assignment is _finalized.
        self.reg_in = next
        net = LogicNet("r", None, args=(self.reg_in,), dests=(self,))
        working_block().add_net(net)


class WrappedWireVector:
    """Wraps a WireVector. Forwards all method calls and attribute accesses.

    WrappedWireVector is useful for dynamically choosing a WireVector base class at
    runtime. If the base class is statically known, do not use WrappedWireVector, and
    just inherit from the base class normally.

    @wire_struct and wire_matrix use WrappedWireVector to implement the
    ``concatenated_type`` option, so an instance can dynamically choose its desired base
    class.
    """

    wire = None

    def __init__(self, wire: WireVector):
        self.__dict__["wire"] = wire

    def __getattr__(self, name: str):
        """Forward all attribute accesses to the wrapped WireVector.

        This does not work for special methods like ``__hash__``. Special methods are
        handled separately below.
        """
        return getattr(self.wire, name)

    def __setattr__(self, name, value):
        """Forward all attribute assignments to the wrapped WireVector.

        This is needed to make ``reg.next <<= foo`` work, because that expands to::

            reg.next = reg.next.__ilshift__(foo)

        See https://docs.python.org/3/library/operator.html#in-place-operators

        This attribute assignment must be forwarded to the underlying Register.
        """
        self.wire.__setattr__(name, value)

    def __hash__(self):
        return hash(self.wire)

    def __str__(self):
        return str(self.wire)

    def __repr__(self):
        return repr(self.wire)

    def __ilshift__(self, other):
        self.wire <<= other
        return self

    def __ior__(self, other):
        self.wire |= other
        return self

    def __bool__(self):
        return bool(self.wire)

    def __and__(self, other):
        return self.wire & other

    def __rand__(self, other):
        return other & self.wire

    def __iand__(self, other):
        self.wire &= other
        return self

    def __or__(self, other):
        return self.wire | other

    def __ror__(self, other):
        return other | self.wire

    def __xor__(self, other):
        return self.wire ^ other

    def __rxor__(self, other):
        return other ^ self.wire

    def __ixor__(self, other):
        self.wire ^= other
        return self

    def __add__(self, other):
        return self.wire + other

    def __radd__(self, other):
        return other + self.wire

    def __iadd__(self, other):
        self.wire += other
        return self

    def __sub__(self, other):
        return self.wire - other

    def __rsub__(self, other):
        return other - self.wire

    def __isub__(self, other):
        self.wire -= other
        return self

    def __mul__(self, other):
        return self.wire * other

    def __rmul__(self, other):
        return other * self.wire

    def __imul__(self, other):
        self.wire *= other
        return self

    def __lt__(self, other):
        return self.wire < other

    def __le__(self, other):
        return self.wire <= other

    def __eq__(self, other):
        return self.wire == other

    def __ne__(self, other):
        return self.wire != other

    def __gt__(self, other):
        return self.wire > other

    def __ge__(self, other):
        return self.wire >= other

    def __invert__(self):
        return ~self.wire

    def __getitem__(self, item):
        return self.wire[item]

    def __lshift__(self, other):
        return self.wire << other

    def __rshift__(self, other):
        return self.wire >> other

    def __mod__(self, other):
        return self.wire % other

    def __len__(self):
        return len(self.wire)

    def __enter__(self):
        self.wire.__enter__()

    def __exit__(self, *execinfo):
        self.wire.__exit__(*execinfo)



# # Render traces with various `WaveRenderer` options.

# Run this demo to see which options work well in your terminal.
import pyrtl


# Make some clock dividers and counters.
def make_clock(period: int):
    """Make a clock signal that inverts every `period` cycles."""
    assert period > 0

    # Build a chain of registers.
    first_reg = pyrtl.Register(bitwidth=1, name=f"clock_0_{period}", reset_value=1)
    last_reg = first_reg
    for offset in range(1, period):
        reg = pyrtl.Register(bitwidth=1, name=f"clock_{offset}_{period}")
        reg.next <<= last_reg
        last_reg = reg

    # The first register's input is the inverse of the last register's output.
    first_reg.next <<= ~last_reg
    return last_reg


def make_counter(period: int, bitwidth: int = 2):
    """Make a counter that increments every `period` cycles."""
    assert period > 0

    # Build a chain of registers.
    first_reg = pyrtl.Register(bitwidth=bitwidth, name=f"counter_0_{period}")
    last_reg = first_reg
    for offset in range(1, period):
        reg = pyrtl.Register(bitwidth=bitwidth, name=f"counter_{offset}_{period}")
        reg.next <<= last_reg
        last_reg = reg

    # The first register's input is the last register's output plus 1.
    first_reg.next <<= last_reg + pyrtl.Const(1)
    return last_reg


make_clock(period=1)
make_clock(period=2)
make_counter(period=1)
make_counter(period=2)

# Simulate 20 cycles.
sim = pyrtl.Simulation()
sim.step_multiple(nsteps=20)

# Render the trace with a variety of rendering options.
renderers = {
    "powerline": (
        pyrtl.simulation.PowerlineRendererConstants(),
        "Requires a font with powerline glyphs",
    ),
    "utf-8": (
        pyrtl.simulation.Utf8RendererConstants(),
        "Unicode, default non-Windows renderer",
    ),
    "utf-8-alt": (
        pyrtl.simulation.Utf8AltRendererConstants(),
        "Unicode, alternate display option",
    ),
    "cp437": (
        pyrtl.simulation.Cp437RendererConstants(),
        "Code page 437 (8-bit ASCII), default Windows renderer",
    ),
    "ascii": (pyrtl.simulation.AsciiRendererConstants(), "Basic 7-bit ASCII renderer"),
}

for name, (constants, notes) in renderers.items():
    print(f"# {notes}")
    print(f"export PYRTL_RENDERER={name}\n")
    sim.tracer.render_trace(
        renderer=pyrtl.simulation.WaveRenderer(constants), repr_func=int
    )
    print()



from __future__ import annotations

import random
from typing import Callable

import pyrtl


def calculate_max_and_min_bitwidths(max_bitwidth=None, exact_bitwidth=None):
    if max_bitwidth is not None:
        min_bitwidth = 1
    elif exact_bitwidth is not None:
        min_bitwidth = max_bitwidth = exact_bitwidth
    else:
        msg = "A max or exact bitwidth must be specified"
        raise pyrtl.PyrtlError(msg)
    return min_bitwidth, max_bitwidth


def inverse_power_dist(bitwidth):
    # Note that this is not uniformly distributed
    return int(2 ** random.uniform(0, bitwidth) - 1)


def uniform_dist(bitwidth):
    return random.randrange(2**bitwidth)


def make_inputs_and_values(
    num_wires: int,
    max_bitwidth: int | None = None,
    exact_bitwidth: int | None = None,
    dist: Callable[[int], int] = uniform_dist,
    test_vals: int = 20,
) -> tuple[list[pyrtl.Input], list[list[int]]]:
    """Generates multiple :class:`.Input` wires and their test values.

    The generated list of test values is a list of lists. The inner lists each represent
    the values of a single :class:`.Input` wire, for each :class:`.Simulation` cycle.

    :param num_wires: Number of :class:`Inputs<.Input>` to generate.
    :param max_bitwidth: If specified, generate :class:`Inputs<.Input>` with random
        :attr:`~.WireVector.bitwidth` in the range ``[1, max_bitwidth)``.
    :param exact_bitwidth: If specified, generate :class:`Inputs<.Input>` with
        :attr:`~.WireVector.bitwidth` ``exact_bitwidth``.
    :param dist: Function to generate the random :class:`.Input` values.
    :param test_vals: Number of random :class:`.Input` values to generate.

    :return: ``(inputs, values)``, where ``inputs`` is a :class:`list` of
             :class:`Inputs<.Input>`, and ``values`` is a list of values for each
             ``input``.
    """
    min_bitwidth, max_bitwidth = calculate_max_and_min_bitwidths(
        max_bitwidth, exact_bitwidth
    )
    wires, vals = list(
        zip(
            *(
                an_input_and_vals(
                    random.randrange(min_bitwidth, max_bitwidth + 1),
                    test_vals,
                    random_dist=dist,
                )
                for i in range(num_wires)
            )
        )
    )
    return wires, vals


def an_input_and_vals(
    bitwidth: int,
    test_vals: int = 20,
    name: str = "",
    random_dist: Callable[[int], int] = uniform_dist,
) -> tuple[pyrtl.Input, list[int]]:
    """Generate an :class:`.Input` wire and random test values for it.

    :param bitwidth: The bitwidth of the random values to generate.
    :param test_vals: Number of random values to generate per :class:`.Input`.
    :param name: Name of the returned :class:`.Input`.

    :return: ``(input_wire, test_values)``, where ``input_wire`` is an :class:`.Input`,
             and ``test_values`` is a list of random test values for ``input_wire``.
    """
    input_wire = pyrtl.Input(bitwidth, name=name)  # Creating a new input wire
    test_vals = [random_dist(bitwidth) for i in range(test_vals)]
    return input_wire, test_vals


# deprecated name
generate_in_wire_and_values = an_input_and_vals


def make_consts(
    num_wires: int,
    max_bitwidth: int | None = None,
    exact_bitwidth: int | None = None,
    random_dist: Callable[[int], int] = inverse_power_dist,
) -> tuple[list[pyrtl.Const], list[int]]:
    """Generate random :class:`.Const` values.

    :param num_wires: Number of :class:`Consts<.Const>` to generate.
    :param max_bitwidth: If specified, generate :class:`Consts<.Const>` with random
        :attr:`~.WireVector.bitwidth` in the range ``[1, max_bitwidth)``.
    :param exact_bitwidth: If specified, generate :class:`Consts<.Const>` with
        :attr:`~.WireVector.bitwidth` ``exact_bitwidth``.
    :param random_dist: Function to generate the random :class:`.Const` values.

    :return: ``(consts, values)``, where ``consts`` is a :class:`list` of
             :class:`Consts<.Const>`, and ``values`` is a list of each ``const``'s
             value.
    """
    min_bitwidth, max_bitwidth = calculate_max_and_min_bitwidths(
        max_bitwidth, exact_bitwidth
    )
    bitwidths = [
        random.randrange(min_bitwidth, max_bitwidth + 1) for i in range(num_wires)
    ]
    wires = [pyrtl.Const(random_dist(b), b) for b in bitwidths]
    vals = [w.val for w in wires]
    return wires, vals


def sim_and_ret_out(
    outwire: pyrtl.WireVector, inwires: list[pyrtl.WireVector], invals: list[list[int]]
) -> list[int]:
    """Run a simulation with ``invals`` for ``inwires`` and return ``outwire``'s values.

    .. WARNING::

        Use :meth:`.Simulation.step_multiple` instead::

            sim = pyrtl.Simulation()
            sim.step_multiple(provided_inputs=dict(zip(inwires, invals)))
            output = sim.tracer.trace[outwire.name]

    :param outwire: The wire to return the values of in each simulation cycle.
    :param inwires: A list of :class:`.Input` wires to provide values for.
    :param invals: A list of :class:`.Input` value lists.

    :return: A list of ``outwire``'s values in each simulation cycle.
    """
    # Pulling the value of outwire straight from the log
    return sim_and_ret_outws(inwires, invals)[outwire.name]


def sim_and_ret_outws(
    inwires: list[pyrtl.WireVector], invals: list[list[int]]
) -> dict[str, list[int]]:
    """Run a simulation with ``invals`` for ``inwires`` and return all wire values.

    .. WARNING::

        Use :meth:`.Simulation.step_multiple` instead::

            sim = pyrtl.Simulation()
            sim.step_multiple(provided_inputs=dict(zip(inwires, invals)))
            outputs = sim.tracer.trace

    :param inwires: A list of :class:`.Input` wires to provide values for.
    :param invals: A list of :class:`.Input` value lists.

    :return: A :class:`dict` mapping from a :class:`WireVector`'s name to a
             :class:`list` of its values in each cycle.
    """
    sim = pyrtl.Simulation()
    sim.step_multiple(provided_inputs=dict(zip(inwires, invals)))
    return sim.tracer.trace


def sim_multicycle(in_dict, hold_dict, hold_cycles, sim=None):
    if sim is None:
        sim = pyrtl.Simulation()
    sim.step(in_dict)
    for _i in range(hold_cycles):
        sim.step(hold_dict)
    return sim.tracer.trace[-1]


def multi_sim_multicycle(in_dict, hold_dict, hold_cycles, sim=None):
    if sim is None:
        sim = pyrtl.Simulation()
    cycles = len(next(iter(in_dict.values())))
    for cycle in range(cycles):
        current_dict = {wire: values[cycle] for wire, values in in_dict}
        cur_result = sim_multicycle(current_dict, hold_dict, hold_cycles, sim)
        if cycle == 0:
            results = {wire: [result_val] for wire, result_val in cur_result}
        else:
            for wire, result_val in cur_result:
                results[wire].append(result_val)
    return results



import glob
import os
import subprocess

import pytest

import pyrtl

"""
Tests all of the files in the example folder

Note that this file is structure dependent, so don't forget to change it if the relative
location of the examples changes
"""


@pytest.mark.parametrize(
    "file", glob.iglob(os.path.dirname(__file__) + "/../examples/*.py")
)
def test_all_examples(file):
    pyrtl.reset_working_block()
    try:
        _ = subprocess.check_output(["python", file])
    except subprocess.CalledProcessError as e:
        raise e



"""
Passes contains prebuilt transformantion passes to do optimization, lowering of the
design to single wire gates (synthesis), along with other ways to change a block.
"""

import collections

from pyrtl import transform  # transform.all_nets looks better than all_nets
from pyrtl.core import (
    Block,
    LogicNet,
    PostSynthBlock,
    _get_debug_mode,
    set_working_block,
    working_block,
)
from pyrtl.corecircuits import (
    _basic_add,
    _basic_eq,
    _basic_gt,
    _basic_lt,
    _basic_mult,
    _basic_select,
    _basic_sub,
    as_wires,
    concat,
    concat_list,
)
from pyrtl.helperfuncs import _NetCount
from pyrtl.memory import MemBlock
from pyrtl.pyrtlexceptions import PyrtlError, PyrtlInternalError
from pyrtl.transform import (
    _get_new_block_mem_instance,
    copy_block,
    net_transform,
    replace_wires,
)
from pyrtl.wire import Const, Input, Output, Register, WireVector

# --------------------------------------------------------------------
#   __   __  ___           __      ___    __
#  /  \ |__)  |  |  |\/| |  /  /\   |  | /  \ |\ |
#  \__/ |     |  |  |  | | /_ /~~\  |  | \__/ | \|
#


def optimize(
    update_working_block: bool = True,
    block: Block = None,
    skip_sanity_check: bool = False,
):
    """Return an optimized version of a synthesized hardware block.

    ``optimize`` works on all hardware designs, both synthesized and non synthesized.

    :param update_working_block: Don't copy the block and optimize the new block
        (defaults to ``True``).
    :param block: The block to optimize (defaults to :ref:`working_block`).
    :param skip_sanity_check: Don't perform :meth:`~Block.sanity_check` on the ``block``
        before, during, and after the optimization passes (defaults to ``False``).
        :meth:`~Block.sanity_check` will always be performed in debug mode
        (:func:`set_debug_mode`).
    """
    block = working_block(block)
    if not update_working_block:
        block = copy_block(block)

    with set_working_block(block, no_sanity_check=True):
        if (not skip_sanity_check) or _get_debug_mode():
            block.sanity_check()
        _remove_wire_nets(block, skip_sanity_check)
        _remove_slice_nets(block, skip_sanity_check)
        constant_propagation(block, True)
        _remove_unlistened_nets(block)
        common_subexp_elimination(block)
        _optimize_inverter_chains(block, skip_sanity_check)
        if (not skip_sanity_check) or _get_debug_mode():
            block.sanity_check()
    return block


def _get_inverter_chains(wire_creator, wire_users):
    """Returns all inverter chains in the block.

    The function returns a list of inverter chains in the block. Each inverter chain is
    represented as a list of the WireVectors in the chain.

    Consider the following circuit, for example::

        A -~-> B -~-> C -w-> X
        D -~-> E -w-> Y

    If the function is called on this circuit, it will return::

        [[A, B, C], [D, E]].
    """

    # Build a list of inverter chains. Each inverter chain is a list of WireVectors,
    # from source to destination.
    inverter_chains = []
    for current_dest, current_creator in wire_creator.items():
        if current_creator.op != "~":
            # Skip non-inverters.
            continue

        # The current inverter connects current_arg (a WireVector) to current_dest (also
        # a WireVector).
        current_arg = current_creator.args[0]
        # current_users is the number of LogicNets that use current_dest.
        current_users = len(wire_users[current_dest])

        # Add the current inverter to the end of this inverter chain.
        append_to = None
        # Add the current inverter to the beginning of this inverter chain.
        prepend_to = None
        next_inverter_chains = []
        for inverter_chain in inverter_chains:
            chain_arg = inverter_chain[0]
            chain_dest = inverter_chain[-1]
            chain_users = len(wire_users[chain_dest])

            if chain_dest is current_arg and chain_users == 1:
                # This chain's only destination is the current inverter. Append the
                # current inverter to the chain.
                append_to = inverter_chain
            elif chain_arg is current_dest and current_users == 1:
                # This chain's only argument is the current inverter. Add the current
                # inverter to the beginning of the chain.
                prepend_to = inverter_chain
            else:
                # The current inverter is not connected to the inverter chain, so we
                # pass the inverter chain through to next_inverter_chains
                next_inverter_chains.append(inverter_chain)

        if append_to and prepend_to:
            # The current inverter joins two existing inverter chains.
            next_inverter_chains.append(append_to + prepend_to)
        elif append_to:
            # Add the current inverter after 'append_to'.
            next_inverter_chains.append([*append_to, current_dest])
        elif prepend_to:
            # Add the current inverter before 'prepend_to'.
            next_inverter_chains.append([current_arg, *prepend_to])
        else:
            # The current inverter is not connected to any inverter chain, so we start a
            # new inverter chain with it
            next_inverter_chains.append([current_arg, current_dest])

        inverter_chains = next_inverter_chains
    return inverter_chains


def _optimize_inverter_chains(block, skip_sanity_check=False):
    """Optimizes inverter chains in the block.

    An inverter chain means two or more inverters directly connected to each other.
    Inverter chains are redundant and can be removed. For example, ``A -~-> B -~-> C
    -w-> X`` can be reduced to ``A -w-> X``.

    After optimization, a chain of an even number of inverters will be reduced a direct
    connection, and a chain of an odd number of inverters will be reduced to one
    inverter.

    If an inverter chain has intermediate users it won't be removed. For example, the
    inverter chain in the following circuit won't be removed::

        A -~-> B -~-> C -w-> X
        B -w-> Y
    """

    # wire_creator maps from WireVector to the LogicNet that defines its value.
    # wire_users maps from WireVector to a list of LogicNets that use its value.
    wire_creator, wire_users = block.net_connections()

    new_logic = set()
    net_removal_set = set()
    wire_removal_set = set()

    # This ProducerList maps the end wire of an inverter chain to its beginning wire. We
    # need this because when removing an inverter chain its end wire gets removed, so we
    # need to replace the source of LogicNets using the end wire of the inverter chain
    # with the chain's beginning wire.
    #
    # We need a ProducerList, rather than a simple dict, because if an inverter chain of
    # more than two inverters has intermediate users, we may have to query the dict
    # multiple times to get the replacement for the inverter chain's last wire. Consider
    # the following circuit, for example:
    #
    # A -~-> B -~-> C -w-> X
    # C -~-> D -~-> E -w-> Y
    #
    # This is the optimized version of the circuit:
    #
    # A -w-> X
    # A -w-> Y
    #
    # The inverter chains found will be A-B-C and C-D-E (two separate chains will be
    # found instead of A-B-C-D-E because C has an intermediate user). In the dict, C
    # will be mapped to A and E will be mapped to C. Hence, when finding the replacement
    # of E, we have to first query the dict to get C, and then query the dict again on C
    # to get A.
    wire_src_dict = _ProducerList()

    for inverter_chain in _get_inverter_chains(wire_creator, wire_users):
        # If len(inverter_chain) = n, there are n-1 inverters in the chain. We only
        # remove inverters if there are at least two inverters in a chain.
        if len(inverter_chain) > 2:
            if (
                len(inverter_chain) % 2 == 1
            ):  # There is an even number of inverters in a chain.
                start_idx = 1
            else:  # There is an odd number of inverters in a chain.
                start_idx = 2
            # Remove wires used in the inverter chain.
            wires_to_remove = inverter_chain[start_idx:]
            wire_removal_set.update(wires_to_remove)
            # Remove inverters used in the chain.
            inverters_to_remove = {wire_creator[wire] for wire in wires_to_remove}
            net_removal_set.update(inverters_to_remove)
            # Map the end wire of the inverter chain to the beginning wire.
            wire_src_dict[inverter_chain[-1]] = inverter_chain[start_idx - 1]

    # This loop recreates the block with inverter chains removed. It adds each LogicNet
    # in the original block to the new block if it is not marked for removal, and
    # replaces the source of the LogicNet if its source was the end wire of a removed
    # inverter chain.
    for net in block.logic:
        if net not in net_removal_set:
            new_logic.add(
                LogicNet(
                    net.op,
                    net.op_param,
                    args=tuple(wire_src_dict.find_producer(x) for x in net.args),
                    dests=net.dests,
                )
            )

    block.logic = new_logic
    for dead_wirevector in wire_removal_set:
        block.remove_wirevector(dead_wirevector)

    if (not skip_sanity_check) or _get_debug_mode():
        block.sanity_check()


class _ProducerList:
    """Maps from wire to its immediate producer and finds ultimate producers."""

    def __init__(self):
        self.dict = {}  # map from wirevector to its direct producer wirevector

    def __getitem__(self, item):
        msg = "You usually don't want the immediate producer"
        raise PyrtlError(msg)

    def __setitem__(self, key, item):
        self.dict[key] = item

    def find_producer(self, item):
        while item in self.dict:
            item = self.dict[item]
        return item


def _remove_wire_nets(block, skip_sanity_check=False):
    """Remove all wire nodes from the block."""

    wire_src_dict = _ProducerList()
    wire_removal_set = set()  # set of all wirevectors to be removed

    # one pass to build the map of value producers and all of the nets and wires to be
    # removed
    for net in block.logic:
        if net.op == "w":
            wire_src_dict[net.dests[0]] = net.args[0]
            if not isinstance(net.dests[0], Output):
                wire_removal_set.add(net.dests[0])

    # second full pass to create the new logic without the wire nets
    new_logic = set()
    for net in block.logic:
        if net.op != "w" or isinstance(net.dests[0], Output):
            new_args = tuple(wire_src_dict.find_producer(x) for x in net.args)
            new_net = LogicNet(net.op, net.op_param, new_args, net.dests)
            new_logic.add(new_net)

    # now update the block with the new logic and remove wirevectors
    block.logic = new_logic
    for dead_wirevector in wire_removal_set:
        block.remove_wirevector(dead_wirevector)

    if (not skip_sanity_check) or _get_debug_mode():
        block.sanity_check()


def _remove_slice_nets(block, skip_sanity_check=False):
    """Remove all unneeded slice nodes from the block.

    Unneeded here means that the source and destination wires of a slice net are exactly
    the same, because the slice takes all the bits, in order, from the source.
    """
    # Turns a net of form on the left into the one on the right:
    #
    #  w1
    #   |
    # [3:0]
    #   |
    # [3:0]     ===>  w1
    #   |             |
    # [3:0] w2      [3:0] w2
    #  / \ /         / \ /
    # ~   +         ~   +
    # |   |         |   |

    wire_src_dict = _ProducerList()
    wire_removal_set = set()  # set of all wirevectors to be removed

    def is_net_slicing_entire_wire(net):
        if net.op != "s":
            return False

        src_wire = net.args[0]
        dst_wire = net.dests[0]
        if len(src_wire) != len(dst_wire):
            return False

        selLower = net.op_param[0]
        selUpper = net.op_param[-1]
        # Check if getting all bits from the src_wire (i.e. consecutive bits, MSB to
        # LSB)
        return net.op_param == tuple(range(selLower, selUpper + 1))

    # one pass to build the map of value producers and all of the nets and wires to be
    # removed
    for net in block.logic:
        if is_net_slicing_entire_wire(net):
            wire_src_dict[net.dests[0]] = net.args[0]
            if not isinstance(net.dests[0], Output):
                wire_removal_set.add(net.dests[0])

    # second full pass to create the new logic without the wire nets
    new_logic = set()
    for net in block.logic:
        if not is_net_slicing_entire_wire(net) or isinstance(net.dests[0], Output):
            new_args = tuple(wire_src_dict.find_producer(x) for x in net.args)
            new_net = LogicNet(net.op, net.op_param, new_args, net.dests)
            new_logic.add(new_net)

    # now update the block with the new logic and remove wirevectors
    block.logic = new_logic
    for dead_wirevector in wire_removal_set:
        del block.wirevector_by_name[dead_wirevector.name]
        block.wirevector_set.remove(dead_wirevector)

    if (not skip_sanity_check) or _get_debug_mode():
        block.sanity_check()


def constant_propagation(block, silence_unexpected_net_warnings=False):
    """Removes excess constants in the block.

    .. note::

        The resulting block can have :class:`WireVectors<WireVector>` that are driven
        but not listened to. These are expected to be removed by
        :func:`_remove_unlistened_nets`
    """
    net_count = _NetCount(block)
    while net_count.shrinking():
        _constant_prop_pass(block, silence_unexpected_net_warnings)


def _constant_prop_pass(block, silence_unexpected_net_warnings=False):
    """Does one constant propagation pass"""
    valid_net_ops = "~&|^nrwcsm@"
    no_optimization_ops = "wcsm@"
    one_var_ops = {
        "~": lambda x, mask: ~x & mask,
        "r": lambda x, _: x,  # This is only valid for constant folding purposes
    }
    two_var_ops = {
        "&": lambda left, right: left & right,
        "|": lambda left, right: left | right,
        "^": lambda left, right: left ^ right,
        "n": lambda left, right: 1 - (left & right),
    }

    def _constant_prop_error(net, error_str):
        if not silence_unexpected_net_warnings:
            msg = f"Unexpected net, {net}, has {error_str}"
            raise PyrtlError(msg)

    def constant_prop_check(net_checking):
        def replace_net(new_net):
            nets_to_remove.add(net_checking)
            nets_to_add.add(new_net)

        def replace_net_with_const(const_val):
            bitwidth = len(net_checking.dests[0])
            new_const_wire = Const(bitwidth=bitwidth, val=const_val, block=block)
            wire_add_set.add(new_const_wire)
            replace_net_with_wire(new_const_wire)

        def replace_net_with_wire(new_wire):
            if isinstance(net_checking.dests[0], Output):
                replace_net(
                    LogicNet("w", None, args=(new_wire,), dests=net_checking.dests)
                )
            else:
                nets_to_remove.add(net_checking)
                new_wire_src[net_checking.dests[0]] = new_wire

        if net_checking.op not in valid_net_ops:
            _constant_prop_error(
                net_checking, "has a net not handled by constant_propagation"
            )
            return  # skip if we are ignoring unoptimizable ops

        num_constants = sum(isinstance(arg, Const) for arg in net_checking.args)

        if num_constants == 0 or net_checking.op in no_optimization_ops:
            return  # assuming wire nets are already optimized

        if (net_checking.op in two_var_ops) and num_constants == 1:
            long_wires = [
                w for w in net_checking.args + net_checking.dests if len(w) != 1
            ]
            if long_wires:
                _constant_prop_error(
                    net_checking,
                    f"has wire(s) {long_wires} with bitwidths that are not 1",
                )
                return  # skip if we are ignoring unoptimizable ops

            # special case
            const_wire, other_wire = net_checking.args
            if isinstance(other_wire, Const):
                const_wire, other_wire = other_wire, const_wire

            outputs = [
                two_var_ops[net_checking.op](const_wire.val, other_val)
                for other_val in (0, 1)
            ]

            if outputs[0] == outputs[1]:
                replace_net_with_const(outputs[0])
            elif outputs[0] == 0:
                replace_net_with_wire(other_wire)
            else:
                replace_net(
                    LogicNet("~", None, args=(other_wire,), dests=net_checking.dests)
                )

        else:
            # this optimization is actually compatible with long wires
            if net_checking.op in two_var_ops:
                output = two_var_ops[net_checking.op](
                    net_checking.args[0].val, net_checking.args[1].val
                )
            else:
                output = one_var_ops[net_checking.op](
                    net_checking.args[0].val, net_checking.args[0].bitmask
                )
            replace_net_with_const(output)

    new_wire_src = _ProducerList()
    wire_add_set = set()
    nets_to_add = set()
    nets_to_remove = set()

    for a_net in block.logic:
        constant_prop_check(a_net)
    # second full pass to cleanup

    new_logic = set()
    for net in block.logic.union(nets_to_add) - nets_to_remove:
        new_args = tuple(new_wire_src.find_producer(x) for x in net.args)
        new_net = LogicNet(net.op, net.op_param, new_args, net.dests)
        new_logic.add(new_net)

    block.logic = new_logic
    for new_wirevector in wire_add_set:
        block.add_wirevector(new_wirevector)

    _remove_unused_wires(block)


def common_subexp_elimination(
    block: Block = None, abs_thresh: float = 1, percent_thresh: float = 0
):
    """Common Subexpression Elimination for PyRTL blocks.

    :param block: the block to run the subexpression elimination on. Defaults to the
        :ref:`working_block`.
    :param abs_thresh: absolute threshold for stopping optimization
    :param percent_thresh: percent threshold for stopping optimization
    """
    block = working_block(block)
    net_count = _NetCount(block)

    while net_count.shrinking(block, percent_thresh, abs_thresh):
        net_table = _find_common_subexps(block)
        _replace_subexps(block, net_table)


ops_where_arg_order_matters = "m@xc<>-"


def _find_common_subexps(block: Block) -> dict[LogicNet, [LogicNet]]:
    """Finds nets that can be considered the same based on op type, op param, and
    arguments.

    Nets are the "same" if:

    1. their op types are the same,

    2. their op_params are the same (e.g. same memory if a memory-related op), and

    3. their arguments are the same (same constant value and bitwidth for const wires,
       otherwise same wire object). The destination wire for a net is not considered.

    :param block: Block to operate over

    :return: mapping from a logic net (with a placehold dest) representing the common
             subexp, to a list of nets matching that common subexp that can be replaced
             with the single common subexp.
    """
    net_table = {}  # {net (without dest) : [net, ...]
    t = ()  # just a placeholder
    const_dict = {}
    for net in block.logic:
        if net.op in ops_where_arg_order_matters:
            new_args = tuple(_const_to_int(w, const_dict) for w in net.args)
        else:
            new_args = tuple(
                sorted((_const_to_int(w, const_dict) for w in net.args), key=hash)
            )
        net_sub = LogicNet(net[0], net[1], new_args, t)  # don't care about dests
        if net_sub in net_table:
            net_table[net_sub].append(net)
        else:
            net_table[net_sub] = [net]
    return net_table


def _const_to_int(wire, const_dict):
    """Return a repr a Const (a tuple composed of width and value) for comparison with
    an 'is'.

    If the wire is not a Const, just return the wire itself; comparison will be done on
    the identity of the wire object instead.
    """
    if isinstance(wire, Const):
        # a very bad hack to make sure two consts will compare
        # correctly with an 'is'
        bitwidth = wire.bitwidth
        val = wire.val
        if bitwidth not in const_dict:
            const_dict[bitwidth] = {val: (bitwidth, val)}
        else:
            if val not in const_dict[bitwidth]:
                const_dict[bitwidth][val] = (bitwidth, val)
        return const_dict[bitwidth][val]

    return wire


def _replace_subexps(block, net_table):
    """Removes unnecessary nets, connecting the common net's dest wire to unnecessary
    net's dest.

    :param block: The block to operate over.
    :param net_table: A mapping from common subexpression (a net) to a list of nets that
        can be replaced with that common net.
    """
    wire_map = {}
    unnecessary_nets = []
    for nets in net_table.values():
        _process_nets_to_discard(nets, wire_map, unnecessary_nets)

    block.logic.difference_update(unnecessary_nets)
    replace_wires(wire_map, block)


def _has_normal_dest_wire(net):
    return not isinstance(net.dests[0], (Register, Output))


def _process_nets_to_discard(nets, wire_map, unnecessary_nets):
    """Helper for tracking how a group of related nets should be replaced with a common
    one.

    :param nets: List of nets that are considered equal and which should be replaced by
        a single common net.
    :param wire_map: Dict that will be updated with a mapping from every old destination
        wire that needs to be removed, to the new destination wire with which it should
        be replaced.
    :param unnecessary_nets: List of nets that are to be discarded.
    """
    if len(nets) == 1:
        return  # also deals with nets with no dest wires
    nets_to_consider = list(filter(_has_normal_dest_wire, nets))

    if len(nets_to_consider) > 1:  # needed to handle cases with only special wires
        net_to_keep = nets_to_consider[0]
        nets_to_discard = nets_to_consider[1:]
        dest_w = net_to_keep.dests[0]
        for net in nets_to_discard:
            old_dst = net.dests[0]
            wire_map[old_dst] = dest_w
            unnecessary_nets.append(net)


def _remove_unlistened_nets(block):
    """Removes all nets that are not connected to an output wirevector.

    :param block: The block to operate over.
    """

    listened_nets = set()
    listened_wires = set()
    prev_listened_net_count = 0

    def add_to_listened(net):
        listened_nets.add(net)
        listened_wires.update(net.args)

    for a_net in block.logic:
        if a_net.op == "@" or any(isinstance(destW, Output) for destW in a_net.dests):
            add_to_listened(a_net)

    while len(listened_nets) > prev_listened_net_count:
        prev_listened_net_count = len(listened_nets)

        for net in block.logic - listened_nets:
            if any((destWire in listened_wires) for destWire in net.dests):
                add_to_listened(net)

    block.logic = listened_nets
    _remove_unused_wires(block)


def _remove_unused_wires(block, keep_inputs=True):
    """Removes all unconnected wires from a block's wirevector_set.

    :param block: The block to operate over.
    :param keep_inputs: If True, retain any Input wires that are not connected to any
        net.
    """
    valid_wires = set()
    for logic_net in block.logic:
        valid_wires.update(logic_net.args, logic_net.dests)

    wire_removal_set = block.wirevector_set.difference(valid_wires)
    for removed_wire in wire_removal_set:
        if isinstance(removed_wire, Input):
            term = " optimized away"
            if keep_inputs:
                valid_wires.add(removed_wire)
                term = " deemed useless by optimization"

            print("Input Wire, " + removed_wire.name + " has been" + term)
        if isinstance(removed_wire, Output):
            PyrtlInternalError("Output wire, " + removed_wire.name + " not driven")

    block.wirevector_set = valid_wires
    block.wirevector_by_name = {wire.name: wire for wire in valid_wires}


# --------------------------------------------------------------------
#    __           ___       ___  __     __
#   /__` \ / |\ |  |  |__| |__  /__` | /__`
#   .__/  |  | \|  |  |  | |___ .__/ | .__/
#


def synthesize(
    update_working_block: bool = True,
    merge_io_vectors: bool = True,
    block: Block = None,
) -> PostSynthBlock:
    """Lower the design to just single-bit "and", "or", "xor", and "not" gates.

    Takes as input a ``block`` (default to :ref:`working_block`) and creates a new block
    which is identical in function but uses only single bit gates and excludes many of
    the more complicated :class:`LogicNet` primitives. The new block should consist
    *almost* exclusively of ``w``, ``&``, ``\\|``, ``^``, and ``~``
    :attr:`ops<LogicNet.op>`, and sequential elements of :class`Registers<Register>`,
    which are one bit as well.

    The two exceptions are for :class:`Inputs<Input>` and :class:`Outputs<Output>`, to
    maintain the same interface, which are immediately broken down into the individual
    bits and memories (read and write ports) which require the reassembly and
    disassembly of the :class:`WireVectors<WireVector>` immediately before and after.
    These are the only two places where ``c`` and ``s`` :attr:`ops<LogicNet.op>` should
    exist. If ``merge_io_vectors`` is ``False``, then these individual bits are not
    reassembled and disassembled before and after, and so no ``c`` and ``s``
    :attr:`ops<LogicNet.op>` will exist. Instead, they will be named ``<name>[n]``,
    where ``n`` is the bit number of original wire to which it corresponds.

    The block that results from synthesis is actually of type :class:`PostSynthBlock`
    which contains a mapping from the original :class:`Inputs<Input>` and
    :class:`Outputs<Output>` to the :class:`Inputs<Input>` and :class:`Outputs<Output>`
    of this block. This is used during :class:`Simulation` to map the
    :class:`Inputs<Input>` and :class:`Outputs<Output>` so that the same testbench can
    be used both pre- and post- synthesis. See documentation for :class:`Simulation` for
    more details.

    :param update_working_block: Boolean specifying if :ref:`working_block` should be
        set to the newly synthesized block.
    :param merge_io_vectors: If ``False``, turn all N-bit IO
        :class:`WireVectors<WireVector>` into N 1-bit IO
        :class:`WireVectors<WireVector>` (i.e. don't maintain interface).
    :param block: The block to synthesize.

    :return: The newly synthesized block, of type :class:`PostSynthBlock`.
    """

    block_pre = working_block(block)
    block_pre.sanity_check()  # before going further, make sure that presynth is valid
    block_in = copy_block(block_pre, update_working_block=False)

    block_out = PostSynthBlock()
    # resulting block should only have one of a restricted set of net ops
    block_out.legal_ops = set("~&|^nrwm@")
    if merge_io_vectors:
        block_out.legal_ops.update(set("cs"))
    wirevector_map = {}  # map from (vector,index) -> new_wire

    with set_working_block(block_out, no_sanity_check=True):
        # First, replace advanced operators with simpler ones
        for op, fun in [
            ("*", _basic_mult),
            ("+", _basic_add),
            ("-", _basic_sub),
            ("x", _basic_select),
            ("=", _basic_eq),
            ("<", _basic_lt),
            (">", _basic_gt),
        ]:
            net_transform(_replace_op(op, fun), block_in)

        # This is a map from the cloned io wirevector created in copy_block,
        # to the original io wirevector found in block_pre. We use it to create
        # the block_out.io_map that is returned to the user.
        orig_io_map = {temp: orig for orig, temp in block_in.io_map.items()}
        orig_reg_map = {temp: orig for orig, temp in block_in.reg_map.items()}

        # Next, create all of the new wires for the new block
        # from the original wires and store them in the wirevector_map
        # for reference.
        for wirevector in block_in.wirevector_subset():
            for i in range(len(wirevector)):
                new_name = "_".join((wirevector.name, "synth", str(i)))
                if isinstance(wirevector, Const):
                    new_val = (wirevector.val >> i) & 0x1
                    new_wirevector = Const(name=new_name, bitwidth=1, val=new_val)
                elif isinstance(wirevector, (Input, Output)):
                    if merge_io_vectors:
                        new_wirevector = WireVector(name="tmp_" + new_name, bitwidth=1)
                    else:
                        # Creating N 1-bit io wires for a given single N-bit io wire.
                        new_name = wirevector.name
                        if len(wirevector) > 1:
                            new_name += "[" + str(i) + "]"
                        new_wirevector = wirevector.__class__(name=new_name, bitwidth=1)
                        block_out.io_map[orig_io_map[wirevector]].append(new_wirevector)
                else:
                    new_wirevector = wirevector.__class__(name=new_name, bitwidth=1)
                    if isinstance(wirevector, Register):
                        block_out.reg_map[orig_reg_map[wirevector]].append(
                            new_wirevector
                        )
                wirevector_map[(wirevector, i)] = new_wirevector

        # Now connect up the inputs and outputs to maintain the interface
        if merge_io_vectors:
            for wirevector in block_in.wirevector_subset(Input):
                input_vector = Input(name=wirevector.name, bitwidth=len(wirevector))
                for i in range(len(wirevector)):
                    wirevector_map[(wirevector, i)] <<= input_vector[i]
                block_out.io_map[orig_io_map[wirevector]].append(input_vector)
            for wirevector in block_in.wirevector_subset(Output):
                output_vector = Output(name=wirevector.name, bitwidth=len(wirevector))
                output_bits = [
                    wirevector_map[(wirevector, i)] for i in range(len(output_vector))
                ]
                output_vector <<= concat_list(output_bits)
                block_out.io_map[orig_io_map[wirevector]].append(output_vector)

        # Now that we have all the wires built and mapped, walk all the blocks
        # and map the logic to the equivalent set of primitives in the system
        out_mems = block_out.mem_map  # dictionary: PreSynth Map -> PostSynth Map
        for net in block_in.logic:
            _decompose(net, wirevector_map, out_mems, block_out)

    if update_working_block:
        set_working_block(block_out, no_sanity_check=True)
    return block_out


def _replace_op(op, fun):
    def _replace_op_inner(net):
        if net.op != op:
            return True
        dest = net.dests[0]
        dest <<= fun(*net.args)
        return False

    return _replace_op_inner


def _decompose(net, wv_map, mems, block_out):
    """Add the wires and logicnets to block_out and wv_map to decompose net"""

    def arg(x, i):
        # return the mapped wire vector for argument x, wire number i
        return wv_map[(net.args[x], i)]

    def destlen():
        # return iterator over length of the destination in bits
        return range(len(net.dests[0]))

    def assign_dest(i, v):
        # assign v to the wiremap for dest[0], wire i
        wv_map[(net.dests[0], i)] <<= v

    one_var_ops = {
        "w": lambda w: w,
        "~": lambda w: ~w,
    }
    c_two_var_ops = {
        "&": lambda left, right: left & right,
        "|": lambda left, right: left | right,
        "^": lambda left, right: left ^ right,
        "n": lambda left, right: left.nand(right),
    }

    if net.op in one_var_ops:
        for i in destlen():
            assign_dest(i, one_var_ops[net.op](arg(0, i)))
    elif net.op in c_two_var_ops:
        for i in destlen():
            assign_dest(i, c_two_var_ops[net.op](arg(0, i), arg(1, i)))
    elif net.op == "s":
        for i in destlen():
            selected_bit = arg(0, net.op_param[i])
            assign_dest(i, selected_bit)
    elif net.op == "c":
        arg_wirelist = []
        # generate list of wires for vectors being concatenated
        for arg_vector in net.args:
            arg_vector_as_list = [
                wv_map[(arg_vector, i)] for i in range(len(arg_vector))
            ]
            arg_wirelist = arg_vector_as_list + arg_wirelist
        for i in destlen():
            assign_dest(i, arg_wirelist[i])
    elif net.op == "r":
        for i in destlen():
            args = (arg(0, i),)
            dests = (wv_map[(net.dests[0], i)],)
            new_net = LogicNet("r", None, args=args, dests=dests)
            block_out.add_net(new_net)
    elif net.op == "m":
        arg0list = [arg(0, i) for i in range(len(net.args[0]))]
        addr = concat_list(arg0list)
        new_mem = _get_new_block_mem_instance(net.op_param, mems, block_out)[1]
        data = as_wires(new_mem[addr])
        for i in destlen():
            assign_dest(i, data[i])
    elif net.op == "@":
        addrlist = [arg(0, i) for i in range(len(net.args[0]))]
        addr = concat_list(addrlist)
        datalist = [arg(1, i) for i in range(len(net.args[1]))]
        data = concat_list(datalist)
        enable = arg(2, 0)
        new_mem = _get_new_block_mem_instance(net.op_param, mems, block_out)[1]
        new_mem[addr] <<= MemBlock.EnabledWrite(data=data, enable=enable)
    else:
        msg = f"Unable to synthesize the following net due to unimplemented op :\n{net}"
        raise PyrtlInternalError(msg)
    return


@transform.all_nets
def nand_synth(net: LogicNet):
    """Synthesizes a :class:`PostSynthBlock` into one consisting of nands and inverters
    in place.

    :param PostSynthBlock block: The block to synthesize.
    """
    if net.op in "~nrwcsm@":
        return True

    def arg(num):
        return net.args[num]

    dest = net.dests[0]
    if net.op == "&":
        dest <<= ~(arg(0).nand(arg(1)))
    elif net.op == "|":
        dest <<= (~arg(0)).nand(~arg(1))
    elif net.op == "^":
        temp_0 = arg(0).nand(arg(1))
        dest <<= temp_0.nand(arg(0)).nand(temp_0.nand(arg(1)))
    else:
        msg = f"Op, '{net.op}' is not supported in nand_synth"
        raise PyrtlError(msg)
    return None


@transform.all_nets
def and_inverter_synth(net: LogicNet):
    """Transforms a decomposed block into one consisting of ands and inverters in place.

    :param Block block: The block to synthesize
    """
    if net.op in "~&rwcsm@":
        return True

    def arg(num):
        return net.args[num]

    dest = net.dests[0]
    if net.op == "|":
        dest <<= ~(~arg(0) & ~arg(1))
    elif net.op == "^":
        all_1 = arg(0) & arg(1)
        all_0 = ~arg(0) & ~arg(1)
        dest <<= ~all_0 & ~all_1
    elif net.op == "n":
        dest <<= ~(arg(0) & arg(1))
    else:
        msg = f"Op, '{net.op}' is not supported in and_inv_synth"
        raise PyrtlError(msg)
    return None


@transform.all_nets
def two_way_concat(net: LogicNet):
    """Transforms a block so all n-way (n > 2) :func:`concats<concat>` are replaced with
    series of 2-way :func:`concats<concat>`.

    This is useful for preparing the netlist for output to other formats, like FIRRTL or
    BTOR2, whose ``concatenate`` operation (``cat`` and ``concat``, respectively), only
    allow two arguments (most-significant wire and least-significant wire).

    :param Block block: The block to transform
    """

    # Turns a netlist of the form (where [] denote nets):
    #
    #  w1  w2  w3
    #   |  |   |
    #   [concat]
    #      |
    #      w4
    #
    # into:
    #
    #  w1 w2   w3
    #   | |    |
    # [concat] |
    #      |   |
    #     [concat]
    #        |
    #      [wire]
    #        |
    #        w4
    if net.op != "c":
        return True

    if len(net.args) <= 2:
        return True

    w = concat(net.args[0], net.args[1])
    for a in net.args[2:]:
        w = concat(w, a)

    dest = net.dests[0]
    dest <<= w
    return None


@transform.all_nets
def one_bit_selects(net: LogicNet):
    """Converts arbitrary-sliced :func:`selects<select>` to concatenations of 1-bit
    :func:`selects<select>`.

    This is useful for preparing the netlist for output to other formats, like FIRRTL or
    BTOR2, whose ``select`` operation (``bits`` and ``slice``, respectively) require
    contiguous ranges.

    Python slices are not necessarily contiguous ranges, e.g. the range ``[::2]``
    (syntactic sugar for ``slice(None, None, 2)``) produces indices ``0``, ``2``, ``4``,
    etc. up to the length of the list on which it is used.

    :param Block block: The block to transform
    """
    if net.op != "s":
        return True

    catlist = [net.args[0][i] for i in net.op_param]
    dest = net.dests[0]
    dest <<= concat_list(catlist)
    return None


def direct_connect_outputs(block=None):
    """Remove 'w' nets immediately before outputs, if possible.

    The 'w' nets that are eligible for removal with this pass meet the following
    requirements:

    - The destination wirevector of the net is an Output

    - The source wirevector of the net doesn't go to any other nets.

    :param block: block to update (defaults to :ref:`working_block`)
    """
    # Turns a netlist of the form (where [] denote nets and o is an Output):
    #
    #  w1 w2
    #   | |
    #   [*]
    #    |
    #    w3
    #    |
    #   [w]
    #    |
    #    o
    #
    # into:
    #
    #  w1 w2
    #   | |
    #   [*]
    #    |
    #    o

    # NOTE: would use transform.all_nets(), but it becomes tricky when we want to remove
    # more than just the current net on a single pass
    block = working_block(block)
    _, dst_nets = block.net_connections()

    nets_to_remove = set()
    nets_to_add = set()
    wirevectors_to_remove = set()

    for net in block.logic:
        if net.op == "@":
            continue

        dest_wire = net.dests[0]
        if dest_wire not in dst_nets or len(dst_nets[dest_wire]) > 1:
            continue

        dst_net = dst_nets[dest_wire][0]
        if dst_net.op != "w" or not isinstance(dst_net.dests[0], Output):
            continue

        new_net = LogicNet(
            op=net.op,
            op_param=net.op_param,
            args=net.args,
            dests=dst_net.dests,
        )
        nets_to_remove.add(net)
        nets_to_remove.add(dst_net)
        wirevectors_to_remove.add(dst_net.args[0])
        nets_to_add.add(new_net)

    block.logic.difference_update(nets_to_remove)
    block.logic.update(nets_to_add)
    for w in wirevectors_to_remove:
        block.remove_wirevector(w)


def _make_tree(wire, block, curr_fanout):
    def f(w, n):
        if n == 1:
            return (w,)
        l_fanout = n // 2
        r_fanout = n - l_fanout
        o = WireVector(len(w), block=block)
        split_net = LogicNet(
            op="w",
            op_param=None,
            args=(w,),
            dests=(o,),
        )
        block.add_net(split_net)
        return f(o, l_fanout) + f(o, r_fanout)

    return f(wire, curr_fanout)


def two_way_fanout(block=None):
    """Update the block such that no wire goes to more than 2 destination nets

    :param block: block to update (defaults to :ref:`working_block`)
    """
    from pyrtl.analysis import fanout

    block = working_block(block)

    _, dst_map = block.net_connections()
    # Two-pass approach: Remember which nets will need to change, in case there are
    # multiple arguments which will be changing along the way.
    nets_to_update = collections.defaultdict(list)
    for wire in block.wirevector_subset(exclude=(Output)):
        curr_fanout = fanout(wire)
        if curr_fanout > 1:
            s = _make_tree(wire, block, curr_fanout)
            curr_ix = 0
            for dst_net in dst_map[wire]:
                for i, arg in enumerate(dst_net.args):
                    if arg is wire:
                        nets_to_update[dst_net].append((wire, i, s[curr_ix]))
                        curr_ix += 1
            if curr_ix != curr_fanout:
                msg = "Calculated fanout does not equal number of wires found"
                raise PyrtlInternalError(msg)

    for old_net, args in nets_to_update.items():

        def get_arg(i, a, args=args):
            for orig, ix, from_tree in args:
                # Checking index as well because the same wire could be used as multiple
                # arguments to the same net.
                if i == ix and a is orig:
                    return from_tree
            return a

        new_net = LogicNet(
            op=old_net.op,
            op_param=old_net.op_param,
            args=tuple(get_arg(ix, a) for ix, a in enumerate(old_net.args)),
            dests=old_net.dests,
        )
        block.add_net(new_net)
        block.logic.remove(old_net)



"""
Transform contains structures helpful for writing analysis and transformation passes
over blocks.

Most of the functions in this module are for advanced users only. However, the following
functions are prebuilt transformations that everyone can use: (As of 7/1/16 there are
none in this folder).

Other user accessible transforms that are based on these function can be found in the
passes module.

PyRTL makes it easy to make your own transformation. However in order to make your first
transform, some knowledge about the structure of PyRTL Internal Representation (IR) of
the circuit is necessary. Specifically, one must know what Block, LogicNet, and
WireVector are as well as how Blocks store the latter two structures (through
Block.logic, Block.wirevector_set, etc).
"""

import functools

from pyrtl.core import LogicNet, set_working_block, working_block
from pyrtl.pyrtlexceptions import PyrtlError
from pyrtl.wire import Const, Input, Output, Register, WireVector


def net_transform(transform_func, block=None, **kwargs):
    """Maps nets to new sets of nets according to a custom function.

    If transform_func does not return ``True``, the original net is removed from the
    block's logic set. The net's argument wire/destination wires are not removed.

    :param transform_func: Function signature: ``func(orig_net (logicnet)) ->
        keep_orig_net (bool)``
    :param block: optional block to work on (defaults to :ref:`working_block`)
    """
    block = working_block(block)
    with set_working_block(block, True):
        for net in block.logic.copy():
            keep_orig_net = transform_func(net, **kwargs)
            if not keep_orig_net:
                block.logic.remove(net)


def all_nets(transform_func):
    """Decorator that wraps a net transform function."""

    @functools.wraps(transform_func)
    def t_res(**kwargs):
        net_transform(transform_func, **kwargs)

    return t_res


def wire_transform(
    transform_func,
    select_types=WireVector,
    exclude_types=(Input, Output, Register, Const),
    block=None,
):
    """Maps Wires to new sets of nets and wires according to a custom function.

    Note that if both new_src and new_dst don't equal orig_wire, orig_wire will be
    removed from the block entirely.

    :param transform_func: The function you want to run on all wires. Function
        signature: func(orig_wire (WireVector)) -> src_wire, dst_wire src_wire is the
        src for the stuff you made in the transform func and dst_wire is the sink. To
        indicate that the wire has not been changed, make src_wire and dst_wire both the
        original wire.
    :param select_types: Type or Tuple of types of WireVectors to replace
    :param exclude_types: Type or Tuple of types of WireVectors to exclude from
        replacement
    :param block: The Block to replace wires on. Defaults to the :ref:`working_block`.
    """
    block = working_block(block)
    src_nets, dst_nets = block.net_connections(include_virtual_nodes=False)
    for orig_wire in block.wirevector_subset(select_types, exclude_types):
        new_src, new_dst = transform_func(orig_wire)
        replace_wire_fast(orig_wire, new_src, new_dst, src_nets, dst_nets, block)


def all_wires(transform_func):
    """Decorator that wraps a wire transform function."""

    @functools.wraps(transform_func)
    def t_res(**kwargs):
        wire_transform(transform_func, **kwargs)

    return t_res


def replace_wires(wire_map, block=None):
    """Replace all wires in a block.

    :param {old_wire: new_wire} wire_map: mapping of old wires to new wires
    :param block: block to operate over (defaults to :ref:`working_block`)
    """
    block = working_block(block)
    src_nets, dst_nets = block.net_connections(include_virtual_nodes=False)
    for old_w, new_w in wire_map.items():
        replace_wire_fast(old_w, new_w, new_w, src_nets, dst_nets, block)


def replace_wire_fast(orig_wire, new_src, new_dst, src_nets, dst_nets, block=None):
    """Replace orig_wire with new_src and/or new_dst.

    The net that orig_wire originates from (its source net) will use new_src as its
    destination wire. The nets that orig_wire went to (its destination nets) will now
    have new_dst as one of their argument wires instead.

    This removes and/or adds nets to the block's logic set. This also *updates* the
    src_nets and dst_nets maps that are passed in, such that the following hold::

        old_src_net = src_nets[orig_wire]
        src_nets[new_src] = old_src_net (where old_src_net.dests = (new_src,))

    and::

        old_dst_nets = dst_nets[orig_wire]
        dst_nets[new_dst] = [old_dst_net (where old_dst_net.args replaces orig_wire with
                                          new_dst) foreach old_dst_net]

    For example, given the graph on left, `replace_wire_fast(w1, w4, w1, ...)` produces
    on right::

        a b c d                   a b    c d
        | | | |                   | |    | |
        net net                   net    net
          | |                      |      |
         w1 w2  ==> produces  ==>  w4 w1 w2
          | |                          | |
          net                          net
           |                            |
           w3                           w3

    And given the graph on the left, `replace_wire_fast(w1, w1, w4, ...)` produces on
    the right::

        a b c d                   a b    c d
        | | | |                   | |    | |
        net net                   net    net
          | |                      |      |
         w1 w2  ==> produces  ==>  w1 w4 w2
          | |                          | |
          net                          net
           |                            |
           w3                           w3

    Calling `replace_wire_fast(w1, w4, w4, ...)`, then, fully replaces w1 with w3 in
    both its argument and dest positions::

        a b c d                   a b c d
        | | | |                   | | | |
        net net                   net net
          | |                      |   |
         w1 w2  ==> produces  ==>  w4 w2
          | |                       | |
          net                       net
           |                         |
           w3                        w3

    :param WireVector orig_wire: Wire to be replaced
    :param WireVector new_src: Wire to replace orig_wire, anywhere orig_wire is the
        destination of a net. Ignored if orig_wire equals new_src.
    :param WireVector new_dst: Wire to replace orig_wire, anywhere orig_wire is an
        argument of a net. Ignored if orig_wire equals new_dst.
    :param {WireVector: LogicNet} src_nets: Maps a wire to the net where it is a dest
    :param {WireVector: List[LogicNet]} dst_nets: Maps a wire to list of nets where it
        is an arg
    :param Block block: The block on which to operate (defaults to :ref:`working_block`)
    """

    def remove_net(net_):
        for arg in set(net_.args):
            dst_nets[arg].remove(net_)
            if not len(dst_nets[arg]):
                del dst_nets[arg]
        if len(net_.dests) == 1:
            del src_nets[net_.dests[0]]
        block.logic.remove(net_)

    def add_net(net_):
        for arg in set(net_.args):
            if arg not in dst_nets:
                dst_nets[arg] = [net_]
            else:
                dst_nets[arg].append(net_)
        if len(net_.dests) == 1:
            src_nets[net_.dests[0]] = net_
        block.add_net(net_)

    # src and dst in this function are all relative to wires
    block = working_block(block)
    if new_src is not orig_wire and orig_wire in src_nets:
        # don't need to add the new_src and new_dst because they were made at creation
        net = src_nets[orig_wire]
        new_net = LogicNet(
            op=net.op,
            op_param=net.op_param,
            args=net.args,
            dests=tuple(new_src if w is orig_wire else w for w in net.dests),
        )
        remove_net(net)
        add_net(new_net)

    if new_dst is not orig_wire and orig_wire in dst_nets:
        old_nets = tuple(
            dst_nets[orig_wire]
        )  # need a copy bc the original will be modified
        for net in old_nets:
            new_net = LogicNet(
                op=net.op,
                op_param=net.op_param,
                dests=net.dests,
                args=tuple(new_dst if w is orig_wire else w for w in net.args),
            )
            remove_net(net)
            add_net(new_net)

    if new_dst is not orig_wire and new_src is not orig_wire:
        block.remove_wirevector(orig_wire)


def clone_wire(old_wire, name=None):
    """Makes a copy of any existing wire.

    This function is mainly intended to be used when the two wires are from different
    blocks. Making two wires with the same name in the same block is not allowed.

    :param old_wire: The wire to clone
    :param name: A name for the new wire (required if the old wire and newly cloned wire
        are part of the same block)
    """
    if name is None:
        if working_block() is old_wire._block:
            msg = (
                "Must provide a name for the newly cloned wire when cloning within the "
                "same block."
            )
            raise PyrtlError(msg)
        name = old_wire.name

    if name in working_block().wirevector_by_name:
        msg = "Cannot give a newly cloned wire the same name as an existing wire."
        raise PyrtlError(msg)

    if isinstance(old_wire, Const):
        return Const(old_wire.val, old_wire.bitwidth, name=name)
    return old_wire.__class__(old_wire.bitwidth, name=name)


def copy_block(block=None, update_working_block=True):
    """Makes a copy of an existing block.

    :param block: The block to clone (defaults to the :ref:`working_block`).

    :return: The resulting block
    """
    block_in = working_block(block)
    block_out, temp_wv_map = _clone_block_and_wires(block_in)
    mems = {}
    for net in block_in.logic:
        _copy_net(block_out, net, temp_wv_map, mems)
    block_out.mem_map = mems
    block_out.io_map = {
        io: w for io, w in temp_wv_map.items() if isinstance(io, (Input, Output))
    }
    block_out.reg_map = {
        r: w for r, w in temp_wv_map.items() if isinstance(r, Register)
    }

    if update_working_block:
        set_working_block(block_out)
    return block_out


def _clone_block_and_wires(block_in):
    """This is a generic function to copy the WireVectors for another round of
    synthesis. This does not split a WireVector with multiple wires.

    :param block_in: The block to change
    :param synth_name: a name to prepend to all new copies of a wire

    :return: the resulting block and a WireVector map
    """
    block_in.sanity_check()  # make sure that everything is valid
    block_out = block_in.__class__()
    temp_wv_map = {}
    with set_working_block(block_out, no_sanity_check=True):
        for wirevector in block_in.wirevector_subset():
            new_wv = clone_wire(wirevector)
            temp_wv_map[wirevector] = new_wv

    return block_out, temp_wv_map


def _copy_net(block_out, net, temp_wv_net, mem_map):
    """This function makes a copy of all nets passed to it for synth uses."""
    new_args = tuple(temp_wv_net[a_arg] for a_arg in net.args)
    new_dests = tuple(temp_wv_net[a_dest] for a_dest in net.dests)
    if net.op in "m@":  # special stuff for copying memories
        new_param = _get_new_block_mem_instance(net.op_param, mem_map, block_out)
    else:
        new_param = net.op_param

    new_net = LogicNet(net.op, new_param, args=new_args, dests=new_dests)
    block_out.add_net(new_net)


def _get_new_block_mem_instance(op_param, mem_map, block_out):
    """Gets the instance of the memory in the new block that is associated with a memory
    in a old block.
    """
    memid, old_mem = op_param
    if old_mem not in mem_map:
        new_mem = old_mem._make_copy(block_out)
        new_mem.id = old_mem.id
        mem_map[old_mem] = new_mem
    return memid, mem_map[old_mem]
